# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from __future__ import annotations

import importlib.util
import os
import re
import sys
import tarfile
import tempfile
import time
import warnings
from collections.abc import Callable
from dataclasses import asdict
from pathlib import Path
from typing import Any

from braket.aws.aws_session import AwsSession
from braket.jobs.config import (
    CheckpointConfig,
    DeviceConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.image_uris import Framework, retrieve_image


def prepare_quantum_job(
    device: str,
    source_module: str,
    entry_point: str | None = None,
    image_uri: str | None = None,
    job_name: str | None = None,
    code_location: str | None = None,
    role_arn: str | None = None,
    hyperparameters: dict[str, Any] | None = None,
    input_data: str | dict | S3DataSourceConfig | None = None,
    instance_config: InstanceConfig | None = None,
    distribution: str | None = None,
    stopping_condition: StoppingCondition | None = None,
    output_data_config: OutputDataConfig | None = None,
    copy_checkpoints_from_job: str | None = None,
    checkpoint_config: CheckpointConfig | None = None,
    aws_session: AwsSession | None = None,
    tags: dict[str, str] | None = None,
    reservation_arn: str | None = None,
) -> dict:
    """Creates a hybrid job by invoking the Braket CreateJob API.

    Args:
        device (str): Device ARN of the QPU device that receives priority quantum
            task queueing once the hybrid job begins running. Each QPU has a separate hybrid jobs
            queue so that only one hybrid job is running at a time. The device string is accessible
            in the hybrid job instance as the environment variable "AMZN_BRAKET_DEVICE_ARN".
            When using embedded simulators, you may provide the device argument as string of the
            form: "local:<provider>/<simulator_name>".

        source_module (str): Path (absolute, relative or an S3 URI) to a python module to be
            tarred and uploaded. If `source_module` is an S3 URI, it must point to a
            tar.gz file. Otherwise, source_module may be a file or directory.

        entry_point (str | None): A str that specifies the entry point of the hybrid job, relative
            to the source module. The entry point must be in the format
            `importable.module` or `importable.module:callable`. For example,
            `source_module.submodule:start_here` indicates the `start_here` function
            contained in `source_module.submodule`. If source_module is an S3 URI,
            entry point must be given. Default: source_module's name

        image_uri (str | None): A str that specifies the ECR image to use for executing the hybrid
            job.`image_uris.retrieve_image()` function may be used for retrieving the ECR image URIs
            for the containers supported by Braket. Default = `<Braket base image_uri>`.

        job_name (str | None): A str that specifies the name with which the hybrid job is created.
            The hybrid job name must be between 0 and 50 characters long and cannot contain
            underscores.
            Default: f'{image_uri_type}-{timestamp}'.

        code_location (str | None): The S3 prefix URI where custom code will be uploaded.
            Default: f's3://{default_bucket_name}/jobs/{job_name}/script'.

        role_arn (str | None): A str providing the IAM role ARN used to execute the
            script. Default: IAM role returned by AwsSession's `get_default_jobs_role()`.

        hyperparameters (dict[str, Any] | None): Hyperparameters accessible to the hybrid job.
            The hyperparameters are made accessible as a Dict[str, str] to the hybrid job.
            For convenience, this accepts other types for keys and values, but `str()`
            is called to convert them before being passed on. Default: None.

        input_data (str | dict | S3DataSourceConfig | None): Information about the training
            data. Dictionary maps channel names to local paths or S3 URIs. Contents found
            at any local paths will be uploaded to S3 at
            f's3://{default_bucket_name}/jobs/{job_name}/data/{channel_name}. If a local
            path, S3 URI, or S3DataSourceConfig is provided, it will be given a default
            channel name "input".
            Default: {}.

        instance_config (InstanceConfig | None): Configuration of the instance(s) for running the
            classical code for the hybrid job. Defaults to
            `InstanceConfig(instanceType='ml.m5.large', instanceCount=1, volumeSizeInGB=30)`.

        distribution (str | None): A str that specifies how the hybrid job should be distributed.
            If set to "data_parallel", the hyperparameters for the hybrid job will be set to use
            data parallelism features for PyTorch or TensorFlow. Default: None.

        stopping_condition (StoppingCondition | None): The maximum length of time, in seconds,
            and the maximum number of quantum tasks that a hybrid job can run before being
            forcefully stopped. Default: StoppingCondition(maxRuntimeInSeconds=5 * 24 * 60 * 60).

        output_data_config (OutputDataConfig | None): Specifies the location for the output of the
            hybrid job.
            Default: OutputDataConfig(s3Path=f's3://{default_bucket_name}/jobs/{job_name}/data',
            kmsKeyId=None).

        copy_checkpoints_from_job (str | None): A str that specifies the hybrid job ARN whose
            checkpoint you want to use in the current hybrid job. Specifying this value will copy
            over the checkpoint data from `use_checkpoints_from_job`'s checkpoint_config s3Uri to
            the current hybrid job's checkpoint_config s3Uri, making it available at
            checkpoint_config.localPath during the hybrid job execution. Default: None

        checkpoint_config (CheckpointConfig | None): Configuration that specifies the location where
            checkpoint data is stored.
            Default: CheckpointConfig(localPath='/opt/jobs/checkpoints',
            s3Uri=f's3://{default_bucket_name}/jobs/{job_name}/checkpoints').

        aws_session (AwsSession | None): AwsSession for connecting to AWS Services.
            Default: AwsSession()

        tags (dict[str, str] | None): Dict specifying the key-value pairs for tagging this
            hybrid job.
            Default: {}.

        reservation_arn (str | None): the reservation window arn provided by Braket
            Direct to reserve exclusive usage for the device to run the hybrid job on.
            Default: None.

    Returns:
        dict: Hybrid job tracking the execution on Amazon Braket.

    Raises:
        ValueError: Raises ValueError if the parameters are not valid.
    """
    param_datatype_map = {
        "instance_config": (instance_config, InstanceConfig),
        "stopping_condition": (stopping_condition, StoppingCondition),
        "output_data_config": (output_data_config, OutputDataConfig),
        "checkpoint_config": (checkpoint_config, CheckpointConfig),
    }

    _validate_params(param_datatype_map)
    aws_session = aws_session or AwsSession()
    device_config = DeviceConfig(device)
    timestamp = str(int(time.time() * 1000))
    job_name = job_name or _generate_default_job_name(image_uri=image_uri, timestamp=timestamp)
    role_arn = role_arn or os.getenv("BRAKET_JOBS_ROLE_ARN", aws_session.get_default_jobs_role())
    hyperparameters = hyperparameters or {}
    hyperparameters = {str(key): str(value) for key, value in hyperparameters.items()}
    input_data = input_data or {}
    tags = tags or {}
    default_bucket = aws_session.default_bucket()
    input_data_list = _process_input_data(input_data, job_name, aws_session, timestamp)
    instance_config = instance_config or InstanceConfig()
    stopping_condition = stopping_condition or StoppingCondition()
    output_data_config = output_data_config or OutputDataConfig()
    checkpoint_config = checkpoint_config or CheckpointConfig()
    code_location = code_location or AwsSession.construct_s3_uri(
        default_bucket,
        "jobs",
        job_name,
        timestamp,
        "script",
    )

    if AwsSession.is_s3_uri(source_module):
        _process_s3_source_module(source_module, entry_point, aws_session, code_location)
    else:
        # if entry point is None, it will be set to default here
        entry_point = _process_local_source_module(
            source_module, entry_point, aws_session, code_location
        )
    algorithm_specification = {
        "scriptModeConfig": {
            "entryPoint": entry_point,
            "s3Uri": f"{code_location}/source.tar.gz",
            "compressionType": "GZIP",
        }
    }
    image_uri = image_uri or retrieve_image(Framework.BASE, aws_session.region)
    algorithm_specification["containerImage"] = {"uri": image_uri}
    if not output_data_config.s3Path:
        output_data_config.s3Path = AwsSession.construct_s3_uri(
            default_bucket,
            "jobs",
            job_name,
            timestamp,
            "data",
        )
    if not checkpoint_config.s3Uri:
        checkpoint_config.s3Uri = AwsSession.construct_s3_uri(
            default_bucket,
            "jobs",
            job_name,
            timestamp,
            "checkpoints",
        )
    if copy_checkpoints_from_job:
        checkpoints_to_copy = aws_session.get_job(copy_checkpoints_from_job)["checkpointConfig"][
            "s3Uri"
        ]
        aws_session.copy_s3_directory(checkpoints_to_copy, checkpoint_config.s3Uri)
    if distribution == "data_parallel":
        distributed_hyperparams = {
            "sagemaker_distributed_dataparallel_enabled": "true",
            "sagemaker_instance_type": instance_config.instanceType,
        }
        hyperparameters.update(distributed_hyperparams)

    create_job_kwargs = {
        "jobName": job_name,
        "roleArn": role_arn,
        "algorithmSpecification": algorithm_specification,
        "inputDataConfig": input_data_list,
        "instanceConfig": asdict(instance_config),
        "outputDataConfig": asdict(output_data_config, dict_factory=_exclude_nones_factory),
        "checkpointConfig": asdict(checkpoint_config),
        "deviceConfig": asdict(device_config),
        "hyperParameters": hyperparameters,
        "stoppingCondition": asdict(stopping_condition),
        "tags": tags,
    }

    if reservation_arn:
        create_job_kwargs.update(
            {
                "associations": [
                    {
                        "arn": reservation_arn,
                        "type": "RESERVATION_TIME_WINDOW_ARN",
                    }
                ]
            }
        )

    return create_job_kwargs


def _generate_default_job_name(
    image_uri: str | None = None, func: Callable | None = None, timestamp: int | str | None = None
) -> str:
    """Generate default job name using the image uri and entrypoint function.

    Args:
        image_uri (str | None): URI for the image container.
        func (Callable | None): The entry point function.
        timestamp (int | str | None): Optional timestamp to use instead of generating one.

    Returns:
        str: Hybrid job name.
    """
    max_length = 50
    timestamp = timestamp if timestamp is not None else str(int(time.time() * 1000))

    if func:
        name = func.__name__.replace("_", "-")
        if len(name) + len(timestamp) > max_length:
            name = name[: max_length - len(timestamp) - 1]
            warnings.warn(
                f"Job name exceeded {max_length} characters. "
                f"Truncating name to {name}-{timestamp}.",
                stacklevel=1,
            )
    elif not image_uri:
        name = "braket-job-default"
    else:
        job_type_match = re.search("/amazon-braket-(.*)-jobs:", image_uri) or re.search(
            "/amazon-braket-([^:/]*)", image_uri
        )
        container = f"-{job_type_match.groups()[0]}" if job_type_match else ""
        name = f"braket-job{container}"
    return f"{name}-{timestamp}"


def _process_s3_source_module(
    source_module: str, entry_point: str, aws_session: AwsSession, code_location: str
) -> None:
    """Check that the source module is an S3 URI of the correct type and that entry point is
    provided.

    Args:
        source_module (str): S3 URI pointing to the tarred source module.
        entry_point (str): Entry point for the hybrid job.
        aws_session (AwsSession): AwsSession to copy source module to code location.
        code_location (str): S3 URI pointing to the location where the code will be
            copied to.

    Raises:
        ValueError: The entry point is None or does not end with .tar.gz.
    """
    if entry_point is None:
        raise ValueError("If source_module is an S3 URI, entry_point must be provided.")
    if not source_module.lower().endswith(".tar.gz"):
        raise ValueError(
            "If source_module is an S3 URI, it must point to a tar.gz file. "
            f"Not a valid S3 URI for parameter `source_module`: {source_module}"
        )
    aws_session.copy_s3_object(source_module, f"{code_location}/source.tar.gz")


def _process_local_source_module(
    source_module: str, entry_point: str, aws_session: AwsSession, code_location: str
) -> str:
    """Check that entry point is valid with respect to source module, or provide a default
    value if entry point is not given. Tar and upload source module to code location in S3.

    Args:
        source_module (str): Local path pointing to the source module.
        entry_point (str): Entry point relative to the source module.
        aws_session (AwsSession): AwsSession for uploading tarred source module.
        code_location (str): S3 URI pointing to the location where the code will
            be uploaded to.

    Raises:
        ValueError: Raised if the source module file is not found.

    Returns:
        str: Entry point.
    """
    try:
        # raises FileNotFoundError if not found
        abs_path_source_module = Path(source_module).resolve(strict=True)
    except FileNotFoundError:
        raise ValueError(f"Source module not found: {source_module}")

    entry_point = entry_point or abs_path_source_module.stem
    _validate_entry_point(abs_path_source_module, entry_point)
    _tar_and_upload_to_code_location(abs_path_source_module, aws_session, code_location)
    return entry_point


def _validate_entry_point(source_module_path: Path, entry_point: str) -> None:
    """Confirm that a valid entry point relative to source module is given.

    Args:
        source_module_path (Path): Path to source module.
        entry_point (str): Entry point relative to source module.

    Raises:
        ValueError: Raised if the module was not found.
    """
    importable, _, _method = entry_point.partition(":")
    sys.path.append(str(source_module_path.parent))
    try:
        # second argument allows relative imports
        importlib.invalidate_caches()
        module = importlib.util.find_spec(importable, source_module_path.stem)
        if module is None:
            raise AssertionError
    # if entry point is nested (ie contains '.'), parent modules are imported
    except (ModuleNotFoundError, AssertionError):
        raise ValueError(f"Entry point module was not found: {importable}")
    finally:
        sys.path.pop()


def _tar_and_upload_to_code_location(
    source_module_path: Path, aws_session: AwsSession, code_location: str
) -> None:
    """Tar and upload source module to code location.

    Args:
        source_module_path (Path): Path to source module.
        aws_session (AwsSession): AwsSession for uploading source module.
        code_location (str): S3 URI pointing to the location where the tarred
            source module will be uploaded to.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        with tarfile.open(f"{temp_dir}/source.tar.gz", "w:gz", dereference=True) as tar:
            tar.add(source_module_path, arcname=source_module_path.name)
        aws_session.upload_to_s3(f"{temp_dir}/source.tar.gz", f"{code_location}/source.tar.gz")


def _validate_params(dict_arr: dict[str, tuple[any, any]]) -> None:
    """Validate that config parameters are of the right type.

    Args:
        dict_arr (dict[str, tuple[any, any]]): dict mapping parameter names to
            a tuple containing the provided value and expected type.

    Raises:
        ValueError: If the user_input is not the same as the expected data type.
    """
    for parameter_name, value_tuple in dict_arr.items():
        user_input, expected_datatype = value_tuple

        if user_input and not isinstance(user_input, expected_datatype):
            raise ValueError(
                f"'{parameter_name}' should be of '{expected_datatype}' "
                f"but user provided {type(user_input)}."
            )


def _process_input_data(
    input_data: str | dict | S3DataSourceConfig,
    job_name: str,
    aws_session: AwsSession,
    subdirectory: str,
) -> list[dict[str, Any]]:
    """Convert input data into a list of dicts compatible with the Braket API.

    Args:
        input_data (str | dict | S3DataSourceConfig): Either a channel definition or a
            dictionary mapping channel names to channel definitions, where a channel definition
            can be an S3DataSourceConfig or a str corresponding to a local prefix or S3 prefix.
        job_name (str): Hybrid job name.
        aws_session (AwsSession): AwsSession for possibly uploading local data.
        subdirectory (str): Subdirectory within job name for S3 locations.

    Returns:
        list[dict[str, Any]]: A list of channel configs.
    """
    if not isinstance(input_data, dict):
        input_data = {"input": input_data}
    for channel_name, data in input_data.items():
        if not isinstance(data, S3DataSourceConfig):
            input_data[channel_name] = _process_channel(
                data, job_name, aws_session, channel_name, subdirectory
            )
    return _convert_input_to_config(input_data)


def _process_channel(
    location: str,
    job_name: str,
    aws_session: AwsSession,
    channel_name: str,
    subdirectory: str,
) -> S3DataSourceConfig:
    """Convert a location to an S3DataSourceConfig, uploading local data to S3, if necessary.

    Args:
        location (str): Local prefix or S3 prefix.
        job_name (str): Hybrid job name.
        aws_session (AwsSession): AwsSession to be used for uploading local data.
        channel_name (str): Name of the channel.
        subdirectory (str): Subdirectory within job name for S3 locations.

    Returns:
        S3DataSourceConfig: S3DataSourceConfig for the channel.
    """
    if AwsSession.is_s3_uri(location):
        return S3DataSourceConfig(location)
    else:
        # local prefix "path/to/prefix" will be mapped to
        # s3://bucket/jobs/job-name/subdirectory/data/input/prefix
        location_name = Path(location).name
        s3_prefix = AwsSession.construct_s3_uri(
            aws_session.default_bucket(),
            "jobs",
            job_name,
            subdirectory,
            "data",
            channel_name,
            location_name,
        )
        aws_session.upload_local_data(location, s3_prefix)
        return S3DataSourceConfig(s3_prefix)


def _convert_input_to_config(input_data: dict[str, S3DataSourceConfig]) -> list[dict[str, Any]]:
    """Convert a dictionary mapping channel names to S3DataSourceConfigs into a list of channel
    configs compatible with the Braket API.

    Args:
        input_data (dict[str, S3DataSourceConfig]): A dictionary mapping channel names to
            S3DataSourceConfig objects.

    Returns:
        list[dict[str, Any]]: A list of channel configs.
    """
    return [
        {
            "channelName": channel_name,
            **data_config.config,
        }
        for channel_name, data_config in input_data.items()
    ]


def _exclude_nones_factory(items: list[tuple]) -> dict:
    return {k: v for k, v in items if v is not None}



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import json

import pytest

from braket.circuits import Operator
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.noise import (
    DampingNoise,
    GeneralizedAmplitudeDampingNoise,
    MultiQubitPauliNoise,
    Noise,
    PauliNoise,
    SingleProbabilisticNoise,
    SingleProbabilisticNoise_34,
    SingleProbabilisticNoise_1516,
)
from braket.circuits.serialization import IRType

invalid_data_qubit_count = [(0, ["foo"])]
invalid_data_ascii_symbols = [(1, None)]
invalid_data_ascii_symbols_length = [(2, ["foo", "boo", "braket"])]
invalid_data_prob = [float("nan"), float("inf"), float("-inf"), 0.95, -2.6]
invalid_data_prob_2 = ["a", 1.0 + 1j]
invalid_data_prob_damping = [float("nan"), float("inf"), float("-inf"), 1.5, -2.6]
invalid_data_prob_damping_2 = ["a", 1.0 + 1j]


@pytest.fixture
def base_noise():
    return Noise(qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def single_probability_noise():
    return SingleProbabilisticNoise(probability=0.1, qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def single_probability_noise_34():
    return SingleProbabilisticNoise_34(probability=0.1, qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def single_probability_noise_1516():
    return SingleProbabilisticNoise_1516(probability=0.1, qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def pauli_noise():
    return PauliNoise(probX=0.1, probY=0.2, probZ=0.3, qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def damping_noise():
    return DampingNoise(gamma=0.2, qubit_count=1, ascii_symbols=["foo"])


@pytest.fixture
def generalized_amplitude_damping_noise():
    return GeneralizedAmplitudeDampingNoise(
        gamma=0.2, probability=0.9, qubit_count=1, ascii_symbols=["foo"]
    )


@pytest.mark.parametrize("qubit_count, ascii_symbols", invalid_data_qubit_count)
def test_invalid_data_qubit_count(qubit_count, ascii_symbols):
    with pytest.raises(ValueError):
        Noise(qubit_count, ascii_symbols)


@pytest.mark.parametrize("qubit_count, ascii_symbols", invalid_data_ascii_symbols)
def test_invalid_data_ascii_symbols(qubit_count, ascii_symbols):
    with pytest.raises(ValueError):
        Noise(qubit_count, ascii_symbols)


@pytest.mark.parametrize("qubit_count, ascii_symbols", invalid_data_ascii_symbols_length)
def test_invalid_data_ascii_symbols_length(qubit_count, ascii_symbols):
    with pytest.raises(ValueError):
        Noise(qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob)
def test_invalid_data_single_prob(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(ValueError):
        SingleProbabilisticNoise(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob)
def test_invalid_data_single_prob_34(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(ValueError):
        SingleProbabilisticNoise_34(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob)
def test_invalid_data_single_prob_1516(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(ValueError):
        SingleProbabilisticNoise_1516(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob_2)
def test_invalid_data_type_single_prob(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(TypeError):
        SingleProbabilisticNoise(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob_2)
def test_invalid_data_type_single_prob_34(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(TypeError):
        SingleProbabilisticNoise_34(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob_2)
def test_invalid_data_type_single_prob_1516(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(TypeError):
        SingleProbabilisticNoise_1516(probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probX", invalid_data_prob)
def test_invalid_data_pauli_probX(probX):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probY = 0.1
    probZ = 0.1
    with pytest.raises(ValueError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probY", invalid_data_prob)
def test_invalid_data_pauli_probY(probY):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probX = 0.1
    probZ = 0.1
    with pytest.raises(ValueError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probZ", invalid_data_prob)
def test_invalid_data_pauli_probZ(probZ):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probX = 0.1
    probY = 0.1
    with pytest.raises(ValueError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probX", invalid_data_prob_2)
def test_invalid_data_type_pauli_probX(probX):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probY = 0.1
    probZ = 0.1
    with pytest.raises(TypeError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probY", invalid_data_prob_2)
def test_invalid_data_type_pauli_probY(probY):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probX = 0.1
    probZ = 0.1
    with pytest.raises(TypeError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probZ", invalid_data_prob_2)
def test_invalid_data_type_pauli_probZ(probZ):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probX = 0.1
    probY = 0.1
    with pytest.raises(TypeError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


def test_invalid_data_pauli_sum():
    qubit_count = 1
    ascii_symbols = ["foo"]
    probX = 0.1
    probY = 0.1
    probZ = 0.9
    with pytest.raises(ValueError):
        PauliNoise(probX, probY, probZ, qubit_count, ascii_symbols)


@pytest.mark.parametrize("gamma", invalid_data_prob_damping)
def test_invalid_data_damping_prob(gamma):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(ValueError):
        DampingNoise(gamma, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob_damping)
def test_invalid_data_generalized_amplitude_damping_prob(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    gamma = 0.1
    with pytest.raises(ValueError):
        GeneralizedAmplitudeDampingNoise(gamma, probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("gamma", invalid_data_prob_damping_2)
def test_invalid_data_type_damping_prob(gamma):
    qubit_count = 1
    ascii_symbols = ["foo"]
    with pytest.raises(TypeError):
        DampingNoise(gamma, qubit_count, ascii_symbols)


@pytest.mark.parametrize("probability", invalid_data_prob_damping_2)
def test_invalid_data_type_generalized_amplitude_damping_prob(probability):
    qubit_count = 1
    ascii_symbols = ["foo"]
    gamma = 0.1
    with pytest.raises(TypeError):
        GeneralizedAmplitudeDampingNoise(gamma, probability, qubit_count, ascii_symbols)


@pytest.mark.parametrize("gamma", invalid_data_prob_damping)
def test_invalid_data_generalized_amplitude_damping_gamma(gamma):
    qubit_count = 1
    ascii_symbols = ["foo"]
    probability = 0.1
    with pytest.raises(ValueError):
        GeneralizedAmplitudeDampingNoise(gamma, probability, qubit_count, ascii_symbols)


def test_ascii_symbols(base_noise):
    assert base_noise.ascii_symbols == ("foo",)


def test_is_operator(base_noise):
    assert isinstance(base_noise, Operator)


@pytest.mark.xfail(raises=NotImplementedError)
def test_to_ir_not_implemented_by_default(base_noise):
    base_noise.to_ir(None)


@pytest.mark.parametrize(
    "ir_type, serialization_properties, expected_exception, expected_message",
    [
        (IRType.JAQCD, None, NotImplementedError, "to_jaqcd has not been implemented yet."),
        (IRType.OPENQASM, None, NotImplementedError, "to_openqasm has not been implemented yet."),
        ("invalid-ir-type", None, ValueError, "Supplied ir_type invalid-ir-type is not supported."),
        (
            IRType.OPENQASM,
            "invalid-serialization-properties",
            ValueError,
            "serialization_properties must be of type OpenQASMSerializationProperties for "
            "IRType.OPENQASM.",
        ),
    ],
)
def test_noise_to_ir(
    ir_type, serialization_properties, expected_exception, expected_message, base_noise
):
    with pytest.raises(expected_exception) as exc:
        base_noise.to_ir(0, ir_type, serialization_properties=serialization_properties)
    assert exc.value.args[0] == expected_message


def test_to_matrix_not_implemented_by_default(base_noise):
    with pytest.raises(NotImplementedError):
        base_noise.to_matrix(None)


def test_invalid_deserializatoin():
    with pytest.raises(NotImplementedError):
        Noise.from_dict({})


@pytest.mark.parametrize(
    "noise, expected_string, expected_repr",
    [
        (Noise(1, ["foo"]), "Noise('qubit_count': 1)", "Noise('qubit_count': 1)"),
        (
            SingleProbabilisticNoise(0.1, 1, ["foo"]),
            "SingleProbabilisticNoise(0.1)",
            "SingleProbabilisticNoise('probability': 0.1, 'qubit_count': 1)",
        ),
        (
            DampingNoise(0.1, 1, ["foo"]),
            "DampingNoise(0.1)",
            "DampingNoise('gamma': 0.1, 'qubit_count': 1)",
        ),
        (
            GeneralizedAmplitudeDampingNoise(0.1, 0.2, 1, ["foo"]),
            "GeneralizedAmplitudeDampingNoise(0.1, 0.2)",
            "GeneralizedAmplitudeDampingNoise('gamma': 0.1, 'probability': 0.2, 'qubit_count': 1)",
        ),
        (
            PauliNoise(0.1, 0.2, 0.3, 1, ["foo"]),
            "PauliNoise(0.1, 0.2, 0.3)",
            "PauliNoise('probX': 0.1, 'probY': 0.2, 'probZ': 0.3, 'qubit_count': 1)",
        ),
        (
            MultiQubitPauliNoise({"X": 0.2}, 1, ["foo"]),
            "MultiQubitPauliNoise({'X': 0.2})",
            "MultiQubitPauliNoise('probabilities' : {'X': 0.2}, 'qubit_count': 1)",
        ),
    ],
)
def test_noise_str_repr(noise, expected_string, expected_repr):
    assert str(noise) == expected_string
    assert repr(noise) == expected_repr


@pytest.mark.parametrize(
    "noise",
    [
        SingleProbabilisticNoise(0.1, 1, ["foo"]),
        DampingNoise(0.1, 1, ["foo"]),
        GeneralizedAmplitudeDampingNoise(0.1, 0.2, 1, ["foo"]),
        PauliNoise(0.1, 0.2, 0.3, 1, ["foo"]),
        MultiQubitPauliNoise({"X": 0.2}, 1, ["foo"]),
    ],
)
def test_noise_serialization(noise):
    representation = noise.to_dict()
    assert isinstance(representation, dict)
    serialized = json.dumps(representation)
    assert isinstance(serialized, str)


@pytest.mark.parametrize(
    "noise, equal_noise, unequal_noise, param_noise",
    [
        (
            SingleProbabilisticNoise(0.1, 1, ["foo"]),
            SingleProbabilisticNoise(0.1, 1, ["foo"]),
            SingleProbabilisticNoise(0.2, 1, ["foo"]),
            SingleProbabilisticNoise(FreeParameter("alpha"), 1, ["foo"]),
        ),
        (
            DampingNoise(0.1, 1, ["foo"]),
            DampingNoise(0.1, 1, ["foo"]),
            DampingNoise(0.2, 1, ["foo"]),
            DampingNoise(FreeParameter("alpha"), 1, ["foo"]),
        ),
        (
            GeneralizedAmplitudeDampingNoise(0.1, 0.2, 1, ["foo"]),
            GeneralizedAmplitudeDampingNoise(0.1, 0.2, 1, ["foo"]),
            GeneralizedAmplitudeDampingNoise(0.2, 0.2, 1, ["foo"]),
            GeneralizedAmplitudeDampingNoise(FreeParameter("alpha"), 0.2, 1, ["foo"]),
        ),
        (
            PauliNoise(0.1, 0.2, 0.3, 1, ["foo"]),
            PauliNoise(0.1, 0.2, 0.3, 1, ["foo"]),
            PauliNoise(0.2, 0.2, 0.3, 1, ["foo"]),
            PauliNoise(FreeParameter("x"), FreeParameter("y"), FreeParameter("z"), 1, ["foo"]),
        ),
        (
            MultiQubitPauliNoise({"X": 0.2}, 1, ["foo"]),
            MultiQubitPauliNoise({"X": 0.2}, 1, ["foo"]),
            MultiQubitPauliNoise({"X": 0.3}, 1, ["foo"]),
            MultiQubitPauliNoise({"X": FreeParameter("alpha")}, 1, ["foo"]),
        ),
    ],
)
def test_noise_equality(noise, equal_noise, unequal_noise, param_noise):
    assert noise == noise
    assert noise is noise
    assert noise == equal_noise
    assert noise is not equal_noise
    assert noise != unequal_noise
    assert noise != param_noise
    assert noise != Noise(qubit_count=1, ascii_symbols=["foo"])


def test_noise_base_not_equal_to_different_type():
    assert Noise(qubit_count=1, ascii_symbols=["foo"]) != "foo"


def test_register_noise():
    class _FooNoise(Noise):
        def __init__(self):
            super().__init__(qubit_count=1, ascii_symbols=["foo"])

    Noise.register_noise(_FooNoise)
    assert Noise._FooNoise().name == _FooNoise().name


@pytest.mark.parametrize(
    "noise_class, params",
    [
        (SingleProbabilisticNoise, {"probability": 0.6}),
        (SingleProbabilisticNoise, {"probability": -0.1}),
        (SingleProbabilisticNoise_34, {"probability": 0.76}),
        (SingleProbabilisticNoise_34, {"probability": -0.1}),
        (SingleProbabilisticNoise_1516, {"probability": 0.93755}),
        (SingleProbabilisticNoise_1516, {"probability": -0.1}),
        (MultiQubitPauliNoise, {"probabilities": {"X": 0.4, "Y": 0.7}}),
        (MultiQubitPauliNoise, {"probabilities": {"X": 0.4, "Y": -0.7}}),
        (PauliNoise, {"probX": 0.5, "probY": 0.5, "probZ": 0.5}),
        (PauliNoise, {"probX": -0.1, "probY": 0, "probZ": 0}),
        (DampingNoise, {"gamma": -0.1}),
        (DampingNoise, {"gamma": 1.1}),
        (GeneralizedAmplitudeDampingNoise, {"gamma": 0.1, "probability": -0.2}),
        (GeneralizedAmplitudeDampingNoise, {"gamma": 0.1, "probability": 1.2}),
    ],
)
def test_invalid_values(noise_class, params):
    with pytest.raises(ValueError):
        noise_class(**params, qubit_count=1, ascii_symbols=["foo"])


@pytest.mark.parametrize(
    "probs, qubit_count, ascii_symbols",
    [
        ({"X": 0.1}, 1, ["PC"]),
        ({"XXY": 0.1}, 3, ["PC3", "PC3", "PC3"]),
        ({"YX": 0.1, "IZ": 0.2}, 2, ["PC2", "PC2"]),
    ],
)
def test_multi_qubit_noise(probs, qubit_count, ascii_symbols):
    noise = MultiQubitPauliNoise(probs, qubit_count, ascii_symbols)
    assert noise.probabilities == probs
    assert noise.qubit_count == qubit_count
    assert noise.ascii_symbols == tuple(ascii_symbols)
    assert noise.parameters == [probs[key] for key in sorted(probs.keys())]


@pytest.mark.xfail(raises=ValueError)
class TestInvalidMultiQubitNoise:
    qubit_count = 1
    ascii_symbols = ["PC2"]

    def test_non_empty(self):
        MultiQubitPauliNoise({}, self.qubit_count, self.ascii_symbols)

    def test_non_identity(self):
        MultiQubitPauliNoise({"I": 0.1}, self.qubit_count, self.ascii_symbols)

    def test_non_equal_length_paulis(self):
        MultiQubitPauliNoise({"X": 0.1, "XY": 0.1}, 1, self.ascii_symbols)
        MultiQubitPauliNoise({"X": 0.1, "Y": 0.1}, 2, ["PC2", "PC2"])

    def test_prob_over_one(self):
        MultiQubitPauliNoise({"X": 0.9, "Y": 0.9}, 1, self.ascii_symbols)
        MultiQubitPauliNoise({"XX": 0.9, "YY": 0.9}, 1, self.ascii_symbols)

    def test_prob_under_one(self):
        MultiQubitPauliNoise({"X": -0.6, "Y": -0.9}, 1, self.ascii_symbols)
        MultiQubitPauliNoise({"XX": -0.9, "YY": -0.9}, 2, ["PC2", "PC2"])

    def test_non_pauli_string(self):
        MultiQubitPauliNoise({"T": 0.1}, 1, self.ascii_symbols)

    def test_individual_probs(self):
        MultiQubitPauliNoise({"X": -0.1}, 1, self.ascii_symbols)
        MultiQubitPauliNoise({"X": 1.1}, 1, self.ascii_symbols)

    @pytest.mark.xfail(raises=TypeError)
    def test_keys_strings(self):
        MultiQubitPauliNoise({1: 1.1}, 1, self.ascii_symbols)

    @pytest.mark.xfail(raises=TypeError)
    def test_values_floats(self):
        MultiQubitPauliNoise({"X": "str"}, 1, self.ascii_symbols)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from copy import deepcopy
from typing import Any

from braket.circuits.gate import Gate
from braket.circuits.serialization import (
    IRType,
    OpenQASMSerializationProperties,
    QubitReferenceType,
)
from braket.pulse.pulse_sequence import PulseSequence
from braket.registers.qubit_set import QubitSet


class GateCalibrations:
    """An object containing gate calibration data. The data respresents the mapping on a particular gate
    on a set of qubits to its calibration to be used by a quantum device. This is represented by a dictionary
    with keys of `Tuple(Gate, QubitSet)` mapped to a `PulseSequence`.
    """  # noqa: E501

    def __init__(
        self,
        pulse_sequences: dict[tuple[Gate, QubitSet], PulseSequence],
    ):
        """Inits a `GateCalibrations`.

        Args:
            pulse_sequences (dict[tuple[Gate, QubitSet], PulseSequence]): A mapping containing a key
                of `(Gate, QubitSet)` mapped to the corresponding pulse sequence.

        """
        self.pulse_sequences: dict[tuple[Gate, QubitSet], PulseSequence] = pulse_sequences

    @property
    def pulse_sequences(self) -> dict[tuple[Gate, QubitSet], PulseSequence]:
        """Gets the mapping of (Gate, Qubit) to the corresponding `PulseSequence`.

        Returns:
            dict[tuple[Gate, QubitSet], PulseSequence]: The calibration data Dictionary.
        """
        return self._pulse_sequences

    @pulse_sequences.setter
    def pulse_sequences(self, value: Any) -> None:
        """Sets the mapping of (Gate, Qubit) to the corresponding `PulseSequence`.

        Args:
            value(Any): The value for the pulse_sequences property to be set to.

        Raises:
            TypeError: Raised if the type is not dict[tuple[Gate, QubitSet], PulseSequence]

        """
        if isinstance(value, dict) and all(
            isinstance(k[0], Gate) and isinstance(k[1], QubitSet) and isinstance(v, PulseSequence)
            for (k, v) in value.items()
        ):
            self._pulse_sequences = value
        else:
            raise TypeError(
                "The value for pulse_sequence must be of type: "
                "dict[tuple[Gate, QubitSet], PulseSequence]"
            )

    def copy(self) -> GateCalibrations:
        """Returns a copy of the object.

        Returns:
            GateCalibrations: a copy of the calibrations.
        """
        return GateCalibrations(deepcopy(self._pulse_sequences))

    def __len__(self):
        return len(self._pulse_sequences)

    def filter(
        self,
        gates: list[Gate] | None = None,
        qubits: QubitSet | list[QubitSet] | None = None,
    ) -> GateCalibrations:
        """Filters the data based on optional lists of gates and QubitSets.

        Args:
            gates (list[Gate] | None): An optional list of gates to filter on.
            qubits (QubitSet | list[QubitSet] | None): An optional `QubitSet` or
                list of `QubitSet` to filter on.

        Returns:
            GateCalibrations: A filtered GateCalibrations object.
        """
        keys = self.pulse_sequences.keys()
        if isinstance(qubits, QubitSet):
            qubits = [qubits]
        filtered_calibration_keys = [
            tup
            for tup in keys
            if (gates is None or tup[0] in gates)
            and (qubits is None or any(qset.issubset(tup[1]) for qset in qubits))
        ]
        return GateCalibrations(
            {k: v for (k, v) in self.pulse_sequences.items() if k in filtered_calibration_keys},
        )

    def to_ir(self, calibration_key: tuple[Gate, QubitSet] | None = None) -> str:
        """Returns the defcal representation for the `GateCalibrations` object.

        Args:
            calibration_key (tuple[Gate, QubitSet] | None): An optional key to get a specific defcal.
                Default: None

        Raises:
            ValueError: Key does not exist in the `GateCalibrations` object.

        Returns:
            str: the defcal string for the object.

        """  # noqa: E501
        if calibration_key is not None:
            if calibration_key not in self.pulse_sequences.keys():
                raise ValueError(
                    f"The key {calibration_key} does not exist in this GateCalibrations object."
                )
            return (
                self.pulse_sequences[calibration_key]
                .to_ir()
                .replace("cal", self._def_cal_gate(calibration_key), 1)
            )
        else:
            defcal = "\n".join(
                v.to_ir().replace("cal", self._def_cal_gate(k), 1)
                for (k, v) in self.pulse_sequences.items()
            )
            return defcal

    def _def_cal_gate(self, gate_key: tuple[Gate, QubitSet]) -> str:
        return " ".join(
            [
                "defcal",
                gate_key[0].to_ir(
                    target=gate_key[1],
                    serialization_properties=OpenQASMSerializationProperties(
                        QubitReferenceType.PHYSICAL
                    ),
                    ir_type=IRType.OPENQASM,
                )[:-1],
            ]
        )

    def __eq__(self, other: GateCalibrations):
        return isinstance(other, GateCalibrations) and other.pulse_sequences == self.pulse_sequences



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from unittest.mock import Mock, call, patch

import pytest

from braket.jobs.metrics_data import MetricsRetrievalError
from braket.jobs.metrics_data.cwl_insights_metrics_fetcher import CwlInsightsMetricsFetcher


@pytest.fixture
def aws_session():
    _aws_session = Mock()
    return _aws_session


EXAMPLE_METRICS_LOG_LINES = [
    [
        {"field": "@timestamp", "value": "Test timestamp 0"},
        {"field": "@message", "value": "Test value 0"},
    ],
    [
        {"field": "@timestamp", "value": "Test timestamp 1"},
        {"field": "@message", "value": "Test value 1"},
    ],
    [
        {"field": "@timestamp", "value": "Test timestamp 2"},
    ],
    [
        {"field": "@message", "value": "Test value 3"},
    ],
    [],
]

EXPECTED_CALL_LIST = [
    call("Test timestamp 0", "Test value 0"),
    call("Test timestamp 1", "Test value 1"),
    call(None, "Test value 3"),
]


@patch("braket.jobs.metrics_data.cwl_insights_metrics_fetcher.LogMetricsParser.get_parsed_metrics")
@patch("braket.jobs.metrics_data.cwl_insights_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_all_metrics_complete_results(mock_add_metrics, mock_get_metrics, aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.start_query.return_value = {"queryId": "test"}
    logs_client_mock.get_query_results.return_value = {
        "status": "Complete",
        "results": EXAMPLE_METRICS_LOG_LINES,
    }
    expected_result = {"Test": [0]}
    mock_get_metrics.return_value = expected_result

    fetcher = CwlInsightsMetricsFetcher(aws_session)

    result = fetcher.get_metrics_for_job("test_job", job_start_time=1, job_end_time=2)
    logs_client_mock.get_query_results.assert_called_with(queryId="test")
    logs_client_mock.start_query.assert_called_with(
        logGroupName="/aws/braket/jobs",
        startTime=1,
        endTime=2,
        queryString="fields @timestamp, @message | filter @logStream like /^test_job\\//"
        " | filter @message like /Metrics - /",
        limit=10000,
    )
    assert mock_add_metrics.call_args_list == EXPECTED_CALL_LIST
    assert result == expected_result


@patch("braket.jobs.metrics_data.cwl_insights_metrics_fetcher.LogMetricsParser.get_parsed_metrics")
@patch("braket.jobs.metrics_data.cwl_insights_metrics_fetcher.LogMetricsParser.parse_log_message")
def test_get_all_metrics_complete_results_stream_prefix(
    mock_add_metrics, mock_get_metrics, aws_session
):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.start_query.return_value = {"queryId": "test"}
    logs_client_mock.get_query_results.return_value = {
        "status": "Complete",
        "results": EXAMPLE_METRICS_LOG_LINES,
    }
    expected_result = {"Test": [0]}
    mock_get_metrics.return_value = expected_result

    fetcher = CwlInsightsMetricsFetcher(aws_session)

    result = fetcher.get_metrics_for_job(
        "test_job", job_start_time=1, job_end_time=2, stream_prefix="test_job/uuid"
    )
    logs_client_mock.get_query_results.assert_called_with(queryId="test")
    logs_client_mock.start_query.assert_called_with(
        logGroupName="/aws/braket/jobs",
        startTime=1,
        endTime=2,
        queryString="fields @timestamp, @message | filter @logStream like /^test_job\\/uuid\\//"
        " | filter @message like /Metrics - /",
        limit=10000,
    )
    assert mock_add_metrics.call_args_list == EXPECTED_CALL_LIST
    assert result == expected_result


def test_get_all_metrics_timeout(aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.start_query.return_value = {"queryId": "test"}
    logs_client_mock.get_query_results.return_value = {"status": "Queued"}

    fetcher = CwlInsightsMetricsFetcher(aws_session, 0.1, 0.2)
    result = fetcher.get_metrics_for_job("test_job")
    logs_client_mock.get_query_results.assert_called()
    assert result == {}


@pytest.mark.xfail(raises=MetricsRetrievalError)
def test_get_all_metrics_failed(aws_session):
    logs_client_mock = Mock()
    aws_session.logs_client = logs_client_mock

    logs_client_mock.start_query.return_value = {"queryId": "test"}
    logs_client_mock.get_query_results.return_value = {"status": "Failed"}

    fetcher = CwlInsightsMetricsFetcher(aws_session)
    fetcher.get_metrics_for_job("test_job")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import numpy as np
import pytest

from braket.ir.ahs import Program
from braket.task_result import (
    AdditionalMetadata,
    AnalogHamiltonianSimulationShotMeasurement,
    AnalogHamiltonianSimulationShotMetadata,
    AnalogHamiltonianSimulationShotResult,
    AnalogHamiltonianSimulationTaskResult,
    TaskMetadata,
)
from braket.tasks import (
    AnalogHamiltonianSimulationQuantumTaskResult,
    AnalogHamiltonianSimulationShotStatus,
)
from braket.tasks.analog_hamiltonian_simulation_quantum_task_result import ShotResult


@pytest.fixture
def task_metadata():
    return TaskMetadata(**{"id": "task_arn", "deviceId": "arn1", "shots": 100})


@pytest.fixture
def additional_metadata():
    return AdditionalMetadata(
        action=Program(
            setup={
                "ahs_register": {
                    "sites": [
                        [0.0, 0.0],
                        [0.0, 3.0e-6],
                        [0.0, 6.0e-6],
                        [3.0e-6, 0.0],
                        [3.0e-6, 3.0e-6],
                        [3.0e-6, 6.0e-6],
                    ],
                    "filling": [1, 1, 1, 1, 0, 0],
                }
            },
            hamiltonian={
                "drivingFields": [
                    {
                        "amplitude": {
                            "time_series": {
                                "values": [0.0, 2.51327e7, 2.51327e7, 0.0],
                                "times": [0.0, 3.0e-7, 2.7e-6, 3.0e-6],
                            },
                            "pattern": "uniform",
                        },
                        "phase": {
                            "time_series": {"values": [0, 0], "times": [0.0, 3.0e-6]},
                            "pattern": "uniform",
                        },
                        "detuning": {
                            "time_series": {
                                "values": [-1.25664e8, -1.25664e8, 1.25664e8, 1.25664e8],
                                "times": [0.0, 3.0e-7, 2.7e-6, 3.0e-6],
                            },
                            "pattern": "uniform",
                        },
                    }
                ],
                "shiftingFields": [
                    {
                        "magnitude": {
                            "time_series": {
                                "values": [-1.25664e8, 1.25664e8],
                                "times": [0.0, 3.0e-6],
                            },
                            "pattern": [0.5, 1.0, 0.5, 0.5, 0.5, 0.5],
                        }
                    }
                ],
            },
        )
    )


@pytest.fixture
def success_measurement():
    return AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Success"),
        shotResult=AnalogHamiltonianSimulationShotResult(
            preSequence=[1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1],
            postSequence=[0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0],
        ),
    )


@pytest.fixture
def partial_success_measurement():
    return AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Partial Success"),
        shotResult=AnalogHamiltonianSimulationShotResult(
            preSequence=[1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1], postSequence=None
        ),
    )


@pytest.fixture
def failed_measurement():
    return AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Failure"),
        shotResult=AnalogHamiltonianSimulationShotResult(preSequence=None, postSequence=None),
    )


@pytest.fixture
def success_measurement_extended():
    return AnalogHamiltonianSimulationShotMeasurement(
        shotMetadata=AnalogHamiltonianSimulationShotMetadata(shotStatus="Success"),
        shotResult=AnalogHamiltonianSimulationShotResult(
            preSequence=[1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1],
            postSequence=[1, 0, 1, 1, 0, 1, 0, 0, 0, 0, 1],
        ),
    )


@pytest.fixture
def measurements(success_measurement, partial_success_measurement, failed_measurement):
    return [success_measurement, partial_success_measurement, failed_measurement]


@pytest.fixture
def measurements_extended(
    success_measurement,
    success_measurement_extended,
):
    return [
        success_measurement,
        success_measurement_extended,
    ]


@pytest.fixture
def result_str_1(task_metadata, additional_metadata, measurements):
    result = AnalogHamiltonianSimulationTaskResult(
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
        measurements=measurements,
    )
    return result.json()


@pytest.fixture
def result_str_2(task_metadata, additional_metadata, measurements):
    result = AnalogHamiltonianSimulationTaskResult(
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
        measurements=None,
    )
    return result.json()


@pytest.fixture
def result_str_3(task_metadata, additional_metadata, measurements_extended):
    result = AnalogHamiltonianSimulationTaskResult(
        taskMetadata=task_metadata,
        additionalMetadata=additional_metadata,
        measurements=measurements_extended,
    )
    return result.json()


def validate_result_from_str_1(result):
    assert len(result.measurements) == 3
    assert result.measurements[0].status == AnalogHamiltonianSimulationShotStatus.SUCCESS
    np.testing.assert_equal(result.measurements[0].pre_sequence, [1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1])
    np.testing.assert_equal(result.measurements[0].post_sequence, [0, 0, 0, 1, 0, 1, 1, 0, 0, 1, 0])
    assert result.measurements[1].status == AnalogHamiltonianSimulationShotStatus.PARTIAL_SUCCESS
    np.testing.assert_equal(result.measurements[1].pre_sequence, [1, 1, 1, 1, 0, 1, 1, 1, 1, 1, 1])
    assert result.measurements[1].post_sequence is None
    assert result.measurements[2].status == AnalogHamiltonianSimulationShotStatus.FAILURE
    assert result.measurements[2].pre_sequence is None
    assert result.measurements[2].post_sequence is None
    assert result.additional_metadata.action is not None


def test_from_object(result_str_1, task_metadata):
    result = AnalogHamiltonianSimulationQuantumTaskResult.from_object(
        AnalogHamiltonianSimulationTaskResult.parse_raw(result_str_1)
    )
    assert result.task_metadata == task_metadata
    validate_result_from_str_1(result)


def test_from_string(result_str_1, task_metadata):
    result = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_1)
    assert result.task_metadata == task_metadata
    validate_result_from_str_1(result)


def test_from_object_equal_to_from_string(result_str_1):
    assert AnalogHamiltonianSimulationQuantumTaskResult.from_object(
        AnalogHamiltonianSimulationTaskResult.parse_raw(result_str_1)
    ) == AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_1)


def test_equality(task_metadata, result_str_1, result_str_2):
    result_1 = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_1)
    result_2 = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_1)
    other_result = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_2)
    non_result = "not a quantum task result"

    assert result_1 == result_2
    assert result_1 is not result_2
    assert result_1 != other_result
    assert result_1 != AnalogHamiltonianSimulationQuantumTaskResult(
        task_metadata=task_metadata,
        additional_metadata=additional_metadata,
        measurements=[result_1.measurements[1], result_1.measurements[0]],
    )
    assert result_1 != non_result


def test_get_counts(result_str_3):
    result = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_3)

    counts = result.get_counts()
    # Partial Success and Failure result status are mapped to counts = None
    expected_counts = {"rrrgeggrrgr": 1, "grggrgrrrrg": 1}
    assert counts == expected_counts


def test_get_counts_failed_task(task_metadata):
    measurement = ShotResult(AnalogHamiltonianSimulationShotStatus.FAILURE, [], [])
    result = AnalogHamiltonianSimulationQuantumTaskResult(
        task_metadata=task_metadata,
        additional_metadata=additional_metadata,
        measurements=[measurement],
    )

    counts = result.get_counts()
    expected_counts = {}
    assert counts == expected_counts


def test_avg_density(result_str_3):
    result = AnalogHamiltonianSimulationQuantumTaskResult.from_string(result_str_3)

    density = result.get_avg_density()
    expected_density = [0.5, 1.0, 0.5, 0.0, 1.0, 0.0, 0.5, 1.0, 1.0, 0.5, 0.5]
    np.testing.assert_almost_equal(density, expected_density)


@pytest.mark.parametrize(
    "shot0, shot1",
    [
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [], []),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [], []),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.FAILURE, [1], [2]),
            ShotResult(AnalogHamiltonianSimulationShotStatus.FAILURE, [1], [2]),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, None, None),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, None, None),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, None, [1, 2]),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, None, [1, 2]),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1, 2], None),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1, 2], None),
        ),
    ],
)
def test_shot_result_equals(shot0, shot1):
    assert shot0 == shot1


@pytest.mark.parametrize(
    "shot0, shot1",
    [
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [], []),
            ShotResult(AnalogHamiltonianSimulationShotStatus.FAILURE, [], []),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1], [2]),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [2], [1]),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1], [2]),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, None, [2]),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1], None),
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1], [2]),
        ),
        (
            ShotResult(AnalogHamiltonianSimulationShotStatus.SUCCESS, [1], None),
            "not a shot",
        ),
    ],
)
def test_shot_result_not_equals(shot0, shot1):
    assert shot0 != shot1



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import pytest

from braket.circuits import FreeParameter, QubitSet
from braket.pulse import (
    ArbitraryWaveform,
    ConstantWaveform,
    DragGaussianWaveform,
    Frame,
    GaussianWaveform,
    Port,
    PulseSequence,
)


@pytest.fixture
def port():
    return Port(port_id="device_port_x0", dt=1e-3, properties={})


@pytest.fixture
def predefined_frame_1(port):
    return Frame(
        frame_id="predefined_frame_1", frequency=2e9, port=port, phase=0, is_predefined=True
    )


@pytest.fixture
def predefined_frame_2(port):
    return Frame(frame_id="predefined_frame_2", frequency=1e6, port=port, is_predefined=True)


@pytest.fixture
def user_defined_frame(port):
    return Frame(
        frame_id="user_defined_frame_0",
        port=port,
        frequency=1e7,
        phase=3.14,
        is_predefined=False,
        properties={"associatedGate": "rz"},
    )


@pytest.fixture
def conflicting_user_defined_frame():
    return Frame(
        frame_id="user_defined_frame_0",
        port=Port("wrong_port", dt=1e-9),
        frequency=1e7,
        phase=3.14,
        is_predefined=False,
        properties={"associatedGate": "rz"},
    )


def test_pulse_sequence_with_user_defined_frame(user_defined_frame):
    pulse_sequence = PulseSequence().set_frequency(user_defined_frame, 6e6)
    expected_str = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    frame user_defined_frame_0 = newframe(device_port_x0, 10000000.0, 3.14);",
            "    set_frequency(user_defined_frame_0, 6000000.0);",
            "}",
        ]
    )
    assert pulse_sequence.to_ir() == expected_str


def test_pulse_sequence_make_bound_pulse_sequence(predefined_frame_1, predefined_frame_2):
    param = FreeParameter("a") + 2 * FreeParameter("b")
    pulse_sequence = (
        PulseSequence()
        .set_frequency(predefined_frame_1, param)
        .shift_frequency(predefined_frame_1, param)
        .set_phase(predefined_frame_1, param)
        .shift_phase(predefined_frame_1, -param)
        .set_scale(predefined_frame_1, param)
        .capture_v0(predefined_frame_1)
        .delay([predefined_frame_1, predefined_frame_2], param)
        .delay(predefined_frame_1, param)
        .delay(predefined_frame_1, 1e-3)
        .barrier([predefined_frame_1, predefined_frame_2])
        .play(
            predefined_frame_1,
            GaussianWaveform(
                length=FreeParameter("length_g"), sigma=FreeParameter("sigma_g"), id="gauss_wf"
            ),
        )
        .play(
            predefined_frame_2,
            DragGaussianWaveform(
                length=FreeParameter("length_dg"),
                sigma=FreeParameter("sigma_dg"),
                beta=0.2,
                id="drag_gauss_wf",
            ),
        )
        .play(
            predefined_frame_1,
            ConstantWaveform(
                length=FreeParameter("length_c"), iq=complex(2, 0.3), id="constant_wf"
            ),
        )
        .play(
            predefined_frame_2,
            ArbitraryWaveform([complex(1, 0.4), 0, 0.3, complex(0.1, 0.2)], id="arb_wf"),
        )
        .capture_v0(predefined_frame_2)
    )
    expected_str_unbound = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    bit[2] psb;",
            *[
                f"    input float {parameter};"
                for parameter in reversed(list(pulse_sequence.parameters))
            ],
            "    waveform gauss_wf = gaussian(length_g * 1s, sigma_g * 1s, 1, false);",
            "    waveform drag_gauss_wf = drag_gaussian(length_dg * 1s,"
            " sigma_dg * 1s, 0.2, 1, false);",
            "    waveform constant_wf = constant(length_c * 1s, 2.0 + 0.3im);",
            "    waveform arb_wf = {1.0 + 0.4im, 0, 0.3, 0.1 + 0.2im};",
            "    set_frequency(predefined_frame_1, a + 2.0 * b);",
            "    shift_frequency(predefined_frame_1, a + 2.0 * b);",
            "    set_phase(predefined_frame_1, a + 2.0 * b);",
            "    shift_phase(predefined_frame_1, -1.0 * a + -2.0 * b);",
            "    set_scale(predefined_frame_1, a + 2.0 * b);",
            "    psb[0] = capture_v0(predefined_frame_1);",
            "    delay[(a + 2.0 * b) * 1s] predefined_frame_1, predefined_frame_2;",
            "    delay[(a + 2.0 * b) * 1s] predefined_frame_1;",
            "    delay[1.0ms] predefined_frame_1;",
            "    barrier predefined_frame_1, predefined_frame_2;",
            "    play(predefined_frame_1, gauss_wf);",
            "    play(predefined_frame_2, drag_gauss_wf);",
            "    play(predefined_frame_1, constant_wf);",
            "    play(predefined_frame_2, arb_wf);",
            "    psb[1] = capture_v0(predefined_frame_2);",
            "}",
        ]
    )
    assert pulse_sequence.to_ir() == expected_str_unbound
    assert pulse_sequence.parameters == {
        FreeParameter("a"),
        FreeParameter("b"),
        FreeParameter("length_g"),
        FreeParameter("length_dg"),
        FreeParameter("sigma_g"),
        FreeParameter("sigma_dg"),
        FreeParameter("length_c"),
    }
    b_bound = pulse_sequence.make_bound_pulse_sequence(
        {"b": 2, "length_g": 1e-3, "length_dg": 3e-3, "sigma_dg": 0.4, "length_c": 4e-3}
    )
    b_bound_call = pulse_sequence(b=2, length_g=1e-3, length_dg=3e-3, sigma_dg=0.4, length_c=4e-3)
    expected_str_b_bound = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    bit[2] psb;",
            *[f"    input float {parameter};" for parameter in reversed(list(b_bound.parameters))],
            "    waveform gauss_wf = gaussian(1.0ms, sigma_g * 1s, 1, false);",
            "    waveform drag_gauss_wf = drag_gaussian(3.0ms, 400.0ms, 0.2, 1, false);",
            "    waveform constant_wf = constant(4.0ms, 2.0 + 0.3im);",
            "    waveform arb_wf = {1.0 + 0.4im, 0, 0.3, 0.1 + 0.2im};",
            "    set_frequency(predefined_frame_1, a + 4.0);",
            "    shift_frequency(predefined_frame_1, a + 4.0);",
            "    set_phase(predefined_frame_1, a + 4.0);",
            "    shift_phase(predefined_frame_1, -1.0 * a + -4.0);",
            "    set_scale(predefined_frame_1, a + 4.0);",
            "    psb[0] = capture_v0(predefined_frame_1);",
            "    delay[(a + 4.0) * 1s] predefined_frame_1, predefined_frame_2;",
            "    delay[(a + 4.0) * 1s] predefined_frame_1;",
            "    delay[1.0ms] predefined_frame_1;",
            "    barrier predefined_frame_1, predefined_frame_2;",
            "    play(predefined_frame_1, gauss_wf);",
            "    play(predefined_frame_2, drag_gauss_wf);",
            "    play(predefined_frame_1, constant_wf);",
            "    play(predefined_frame_2, arb_wf);",
            "    psb[1] = capture_v0(predefined_frame_2);",
            "}",
        ]
    )
    assert b_bound.to_ir() == b_bound_call.to_ir() == expected_str_b_bound
    assert pulse_sequence.to_ir() == expected_str_unbound
    assert b_bound.parameters == {FreeParameter("sigma_g"), FreeParameter("a")}
    both_bound = b_bound.make_bound_pulse_sequence({"a": 1, "sigma_g": 0.7})
    both_bound_call = b_bound_call(1, sigma_g=0.7)  # use arg 1 for a
    expected_str_both_bound = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    bit[2] psb;",
            "    waveform gauss_wf = gaussian(1.0ms, 700.0ms, 1, false);",
            "    waveform drag_gauss_wf = drag_gaussian(3.0ms, 400.0ms, 0.2, 1, false);",
            "    waveform constant_wf = constant(4.0ms, 2.0 + 0.3im);",
            "    waveform arb_wf = {1.0 + 0.4im, 0, 0.3, 0.1 + 0.2im};",
            "    set_frequency(predefined_frame_1, 5.0);",
            "    shift_frequency(predefined_frame_1, 5.0);",
            "    set_phase(predefined_frame_1, 5.0);",
            "    shift_phase(predefined_frame_1, -5.0);",
            "    set_scale(predefined_frame_1, 5.0);",
            "    psb[0] = capture_v0(predefined_frame_1);",
            "    delay[5.0s] predefined_frame_1, predefined_frame_2;",
            "    delay[5.0s] predefined_frame_1;",
            "    delay[1.0ms] predefined_frame_1;",
            "    barrier predefined_frame_1, predefined_frame_2;",
            "    play(predefined_frame_1, gauss_wf);",
            "    play(predefined_frame_2, drag_gauss_wf);",
            "    play(predefined_frame_1, constant_wf);",
            "    play(predefined_frame_2, arb_wf);",
            "    psb[1] = capture_v0(predefined_frame_2);",
            "}",
        ]
    )
    assert both_bound.to_ir() == both_bound_call.to_ir() == expected_str_both_bound
    assert b_bound.to_ir() == b_bound_call.to_ir() == expected_str_b_bound
    assert pulse_sequence.to_ir() == expected_str_unbound


@pytest.mark.parametrize(
    "method_name, method_kwargs",
    [
        ("set_frequency", {"frequency": 1e4}),
        ("shift_frequency", {"frequency": 2e4}),
        ("set_phase", {"phase": 0.3}),
        ("shift_phase", {"phase": 0.2}),
        ("set_scale", {"scale": 0.7}),
        ("delay", {"duration": 1e-5}),
        ("barrier", None),
        ("play", {"waveform": ConstantWaveform(1e-3, complex(1, 2), "wf_id")}),
    ],
)
def test_pulse_sequence_conflicting_frames(
    user_defined_frame, conflicting_user_defined_frame, method_name, method_kwargs
):
    ps = PulseSequence().shift_frequency(user_defined_frame, 1e2)
    method = getattr(ps, method_name)

    with pytest.raises(ValueError):
        if method_kwargs:
            method(conflicting_user_defined_frame, **method_kwargs)
        else:
            method(conflicting_user_defined_frame)


def test_pulse_sequence_conflicting_wf(user_defined_frame):
    wf = ConstantWaveform(1e-3, complex(1, 2), "wf_id")
    conflicting_wf = ConstantWaveform(1e-4, complex(1, 2), "wf_id")
    ps = PulseSequence().play(user_defined_frame, wf)

    with pytest.raises(ValueError):
        ps.play(user_defined_frame, conflicting_wf)


def test_pulse_sequence_to_time_trace_program_mutation(user_defined_frame):
    wf = ConstantWaveform(1e-3, complex(1, 2), "wf_id")
    ps = PulseSequence().play(user_defined_frame, wf)
    initial_vars = dict(ps._program.undeclared_vars)
    ps.to_time_trace()
    assert initial_vars == ps._program.undeclared_vars


def test_pulse_sequence_to_ir(predefined_frame_1, predefined_frame_2):
    pulse_sequence = (
        PulseSequence()
        .set_frequency(predefined_frame_1, 3e9)
        .shift_frequency(predefined_frame_1, 1e9)
        .set_phase(predefined_frame_1, -0.5)
        .shift_phase(predefined_frame_1, 0.1)
        .set_scale(predefined_frame_1, 0.25)
        .capture_v0(predefined_frame_1)
        .delay([predefined_frame_1, predefined_frame_2], 2e-9)
        .delay(predefined_frame_1, 1e-6)
        .delay(QubitSet(0), 1e-3)
        .barrier(QubitSet([0, 1]))
        .barrier([predefined_frame_1, predefined_frame_2])
        .play(predefined_frame_1, GaussianWaveform(length=1e-3, sigma=0.7, id="gauss_wf"))
        .play(
            predefined_frame_2,
            DragGaussianWaveform(length=3e-3, sigma=0.4, beta=0.2, id="drag_gauss_wf"),
        )
        .play(
            predefined_frame_1,
            ConstantWaveform(length=4e-3, iq=complex(2, 0.3), id="constant_wf"),
        )
        .play(
            predefined_frame_2,
            ArbitraryWaveform([complex(1, 0.4), 0, 0.3, complex(0.1, 0.2)], id="arb_wf"),
        )
        .capture_v0(predefined_frame_2)
    )
    expected_str = "\n".join(
        [
            "OPENQASM 3.0;",
            "cal {",
            "    bit[2] psb;",
            "    waveform gauss_wf = gaussian(1.0ms, 700.0ms, 1, false);",
            "    waveform drag_gauss_wf = drag_gaussian(3.0ms, 400.0ms, 0.2, 1, false);",
            "    waveform constant_wf = constant(4.0ms, 2.0 + 0.3im);",
            "    waveform arb_wf = {1.0 + 0.4im, 0, 0.3, 0.1 + 0.2im};",
            "    set_frequency(predefined_frame_1, 3000000000.0);",
            "    shift_frequency(predefined_frame_1, 1000000000.0);",
            "    set_phase(predefined_frame_1, -0.5);",
            "    shift_phase(predefined_frame_1, 0.1);",
            "    set_scale(predefined_frame_1, 0.25);",
            "    psb[0] = capture_v0(predefined_frame_1);",
            "    delay[2.0ns] predefined_frame_1, predefined_frame_2;",
            "    delay[1.0us] predefined_frame_1;",
            "    delay[1.0ms] $0;",
            "    barrier $0, $1;",
            "    barrier predefined_frame_1, predefined_frame_2;",
            "    play(predefined_frame_1, gauss_wf);",
            "    play(predefined_frame_2, drag_gauss_wf);",
            "    play(predefined_frame_1, constant_wf);",
            "    play(predefined_frame_2, arb_wf);",
            "    psb[1] = capture_v0(predefined_frame_2);",
            "}",
        ]
    )
    assert pulse_sequence.to_ir() == expected_str


def test_parse_from_calibration_schema(predefined_frame_1, predefined_frame_2):
    waveforms = {
        "drag_gauss_wf": DragGaussianWaveform(length=3e-3, sigma=0.4, beta=0.2, id="drag_gauss_wf")
    }
    frames = {predefined_frame_1.id: predefined_frame_1, predefined_frame_2.id: predefined_frame_2}

    calibration_instrs = [
        {
            "name": "barrier",
            "arguments": [{"name": "qubit", "value": "0", "type": "string"}],
        },
        {
            "name": "play",
            "arguments": [
                {"name": "frame", "value": "predefined_frame_1", "type": "frame"},
                {
                    "name": "waveform",
                    "value": "drag_gauss_wf",
                    "type": "waveform",
                },
            ],
        },
        {
            "name": "barrier",
            "arguments": [
                {"name": "frame", "value": "predefined_frame_1", "type": "frame"},
                {"name": "frame", "value": "predefined_frame_2", "type": "frame"},
            ],
        },
        {
            "name": "barrier",
            "arguments": None,
        },
        {
            "name": "delay",
            "arguments": [
                {"name": "duration", "value": 3e-07, "type": "float"},
                {"name": "qubit", "value": "0", "type": "string"},
                {"name": "qubit", "value": "1", "type": "string"},
            ],
        },
        {
            "name": "delay",
            "arguments": [
                {"name": "frame", "value": "predefined_frame_1", "type": "frame"},
                {"name": "duration", "value": 3e-07, "type": "float"},
            ],
        },
        {
            "name": "shift_phase",
            "arguments": [
                {"name": "frame", "value": "predefined_frame_1", "type": "frame"},
                {"name": "phase", "value": 3e-07, "type": "float"},
            ],
        },
        {
            "name": "shift_frequency",
            "arguments": [
                {"name": "frequency", "value": "theta", "type": "expr"},
                {"name": "frame", "value": "predefined_frame_1", "type": "frame"},
                {"name": "extra", "value": "predefined_frame_1", "type": "string"},
            ],
        },
    ]

    expected_pulse_sequence = (
        PulseSequence()
        .barrier(QubitSet(0))
        .play(predefined_frame_1, waveforms["drag_gauss_wf"])
        .barrier([predefined_frame_1, predefined_frame_2])
        .barrier([])
        .delay(QubitSet([0, 1]), 3e-07)
        .delay([predefined_frame_1], 3e-07)
        .shift_phase(predefined_frame_1, 3e-07)
        .shift_frequency(predefined_frame_1, FreeParameter("theta"))
    )

    assert (
        PulseSequence._parse_from_calibration_schema(calibration_instrs, waveforms, frames)
        == expected_pulse_sequence
    )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import math
import tarfile
import tempfile
import time
from enum import Enum
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, ClassVar

import boto3
from botocore.exceptions import ClientError

from braket.aws import AwsDevice
from braket.aws.aws_session import AwsSession
from braket.aws.queue_information import HybridJobQueueInfo
from braket.jobs import logs
from braket.jobs.config import (
    CheckpointConfig,
    InstanceConfig,
    OutputDataConfig,
    S3DataSourceConfig,
    StoppingCondition,
)
from braket.jobs.data_persistence import load_job_result
from braket.jobs.metrics_data.cwl_insights_metrics_fetcher import CwlInsightsMetricsFetcher

# TODO: Have added metric file in metrics folder, but have to decide on the name for keep
# for the files, since all those metrics are retrieved from the CW.
from braket.jobs.metrics_data.definitions import MetricStatistic, MetricType
from braket.jobs.quantum_job import QuantumJob
from braket.jobs.quantum_job_creation import prepare_quantum_job


class AwsQuantumJob(QuantumJob):
    """Amazon Braket implementation of a quantum job."""

    TERMINAL_STATES: ClassVar[set[str]] = {"CANCELLED", "COMPLETED", "FAILED"}
    RESULTS_FILENAME = "results.json"
    RESULTS_TAR_FILENAME = "model.tar.gz"
    LOG_GROUP = "/aws/braket/jobs"

    class LogState(Enum):
        """Log state enum."""

        TAILING = "tailing"
        JOB_COMPLETE = "job_complete"
        COMPLETE = "complete"

    @classmethod
    def create(
        cls,
        device: str,
        source_module: str,
        entry_point: str | None = None,
        image_uri: str | None = None,
        job_name: str | None = None,
        code_location: str | None = None,
        role_arn: str | None = None,
        wait_until_complete: bool = False,
        hyperparameters: dict[str, Any] | None = None,
        input_data: str | dict | S3DataSourceConfig | None = None,
        instance_config: InstanceConfig | None = None,
        distribution: str | None = None,
        stopping_condition: StoppingCondition | None = None,
        output_data_config: OutputDataConfig | None = None,
        copy_checkpoints_from_job: str | None = None,
        checkpoint_config: CheckpointConfig | None = None,
        aws_session: AwsSession | None = None,
        tags: dict[str, str] | None = None,
        logger: Logger = getLogger(__name__),
        quiet: bool = False,
        reservation_arn: str | None = None,
    ) -> AwsQuantumJob:
        """Creates a hybrid job by invoking the Braket CreateJob API.

        Args:
            device (str): Device ARN of the QPU device that receives priority quantum
                task queueing once the hybrid job begins running. Each QPU has a separate hybrid
                jobs queue so that only one hybrid job is running at a time. The device string is
                accessible in the hybrid job instance as the environment variable
                "AMZN_BRAKET_DEVICE_ARN". When using embedded simulators, you may provide the device
                argument as a string of the form: "local:<provider>/<simulator_name>".

            source_module (str): Path (absolute, relative or an S3 URI) to a python module to be
                tarred and uploaded. If `source_module` is an S3 URI, it must point to a
                tar.gz file. Otherwise, source_module may be a file or directory.

            entry_point (str | None): A str that specifies the entry point of the hybrid job,
                relative to the source module. The entry point must be in the format
                `importable.module` or `importable.module:callable`. For example,
                `source_module.submodule:start_here` indicates the `start_here` function
                contained in `source_module.submodule`. If source_module is an S3 URI,
                entry point must be given. Default: source_module's name

            image_uri (str | None): A str that specifies the ECR image to use for executing the
                hybrid job. `image_uris.retrieve_image()` function may be used for retrieving the
                ECR image URIs for the containers supported by Braket.
                Default = `<Braket base image_uri>`.

            job_name (str | None): A str that specifies the name with which the hybrid job is
                created. Allowed pattern for hybrid job name: `^[a-zA-Z0-9](-*[a-zA-Z0-9]){0,50}$`
                Default: f'{image_uri_type}-{timestamp}'.

            code_location (str | None): The S3 prefix URI where custom code will be uploaded.
                Default: f's3://{default_bucket_name}/jobs/{job_name}/script'.

            role_arn (str | None): A str providing the IAM role ARN used to execute the
                script. Default: IAM role returned by AwsSession's `get_default_jobs_role()`.

            wait_until_complete (bool): `True` if we should wait until the hybrid job completes.
                This would tail the hybrid job logs as it waits. Otherwise `False`.
                Default: `False`.

            hyperparameters (dict[str, Any] | None): Hyperparameters accessible to the hybrid job.
                The hyperparameters are made accessible as a dict[str, str] to the hybrid job.
                For convenience, this accepts other types for keys and values, but `str()`
                is called to convert them before being passed on. Default: None.

            input_data (str | dict | S3DataSourceConfig | None): Information about the training
                data. Dictionary maps channel names to local paths or S3 URIs. Contents found
                at any local paths will be uploaded to S3 at
                f's3://{default_bucket_name}/jobs/{job_name}/data/{channel_name}. If a local
                path, S3 URI, or S3DataSourceConfig is provided, it will be given a default
                channel name "input".
                Default: {}.

            instance_config (InstanceConfig | None): Configuration of the instance(s) for running
                the classical code for the hybrid job. Default:
                `InstanceConfig(instanceType='ml.m5.large', instanceCount=1, volumeSizeInGB=30)`.

            distribution (str | None): A str that specifies how the hybrid job should be
                distributed. If set to "data_parallel", the hyperparameters for the hybrid job will
                be set to use data parallelism features for PyTorch or TensorFlow. Default: None.

            stopping_condition (StoppingCondition | None): The maximum length of time, in seconds,
                and the maximum number of quantum tasks that a hybrid job can run before being
                forcefully stopped.
                Default: StoppingCondition(maxRuntimeInSeconds=5 * 24 * 60 * 60).

            output_data_config (OutputDataConfig | None): Specifies the location for the output of
                the hybrid job.
                Default: OutputDataConfig(s3Path=f's3://{default_bucket_name}/jobs/{job_name}/data',
                kmsKeyId=None).

            copy_checkpoints_from_job (str | None): A str that specifies the hybrid job ARN whose
                checkpoint you want to use in the current hybrid job. Specifying this value will
                copy over the checkpoint data from `use_checkpoints_from_job`'s checkpoint_config
                s3Uri to the current hybrid job's checkpoint_config s3Uri, making it available at
                checkpoint_config.localPath during the hybrid job execution. Default: None

            checkpoint_config (CheckpointConfig | None): Configuration that specifies the location
                where checkpoint data is stored.
                Default: CheckpointConfig(localPath='/opt/jobs/checkpoints',
                s3Uri=f's3://{default_bucket_name}/jobs/{job_name}/checkpoints').

            aws_session (AwsSession | None): AwsSession for connecting to AWS Services.
                Default: AwsSession()

            tags (dict[str, str] | None): Dict specifying the key-value pairs for tagging this
                hybrid job.
                Default: {}.

            logger (Logger): Logger object with which to write logs, such as quantum task statuses
                while waiting for quantum task to be in a terminal state. Default is
                `getLogger(__name__)`

            quiet (bool): Sets the verbosity of the logger to low and does not report queue
                position. Default is `False`.

            reservation_arn (str | None): the reservation window arn provided by Braket
                Direct to reserve exclusive usage for the device to run the hybrid job on.
                Default: None.

        Returns:
            AwsQuantumJob: Hybrid job tracking the execution on Amazon Braket.

        Raises:
            ValueError: Raises ValueError if the parameters are not valid.
        """
        aws_session = AwsQuantumJob._initialize_session(aws_session, device, logger)

        create_job_kwargs = prepare_quantum_job(
            device=device,
            source_module=source_module,
            entry_point=entry_point,
            image_uri=image_uri,
            job_name=job_name,
            code_location=code_location,
            role_arn=role_arn,
            hyperparameters=hyperparameters,
            input_data=input_data,
            instance_config=instance_config,
            distribution=distribution,
            stopping_condition=stopping_condition,
            output_data_config=output_data_config,
            copy_checkpoints_from_job=copy_checkpoints_from_job,
            checkpoint_config=checkpoint_config,
            aws_session=aws_session,
            tags=tags,
            reservation_arn=reservation_arn,
        )

        job_arn = aws_session.create_job(**create_job_kwargs)
        job = AwsQuantumJob(job_arn, aws_session, quiet)

        if wait_until_complete:
            print(f"Initializing Braket Job: {job_arn}")
            job.logs(wait=True)

        return job

    def __init__(self, arn: str, aws_session: AwsSession | None = None, quiet: bool = False):
        """Initializes an `AwsQuantumJob`.

        Args:
            arn (str): The ARN of the hybrid job.
            aws_session (AwsSession | None): The `AwsSession` for connecting to AWS services.
                Default is `None`, in which case an `AwsSession` object will be created with the
                region of the hybrid job.
            quiet (bool): Sets the verbosity of the logger to low and does not report queue
                position. Default is `False`.

        Raises:
            ValueError: Supplied region and session region do not match.
        """
        self._arn: str = arn
        self._quiet = quiet
        if aws_session:
            if not self._is_valid_aws_session_region_for_job_arn(aws_session, arn):
                raise ValueError(
                    "The aws session region does not match the region for the supplied arn."
                )
            self._aws_session = aws_session
        else:
            self._aws_session = AwsQuantumJob._default_session_for_job_arn(arn)
        self._metadata = {}

    @staticmethod
    def _is_valid_aws_session_region_for_job_arn(aws_session: AwsSession, job_arn: str) -> bool:
        """Checks whether the job region and session region match.

        Returns:
            bool: `True` when the aws_session region matches the job_arn region; otherwise
            `False`.
        """
        job_region = job_arn.split(":")[3]
        return job_region == aws_session.region

    @staticmethod
    def _default_session_for_job_arn(job_arn: str) -> AwsSession:
        """Get an AwsSession for the Hybrid Job ARN. The AWS session should be in the region of the
           hybrid job.

        Args:
            job_arn (str): The ARN for the quantum hybrid job.

        Returns:
            AwsSession: `AwsSession` object with default `boto_session` in hybrid job's region.
        """
        job_region = job_arn.split(":")[3]
        boto_session = boto3.Session(region_name=job_region)
        return AwsSession(boto_session=boto_session)

    @property
    def arn(self) -> str:
        """str: The ARN (Amazon Resource Name) of the quantum hybrid job."""
        return self._arn

    @property
    def name(self) -> str:
        """str: The name of the quantum job."""
        return self.metadata(use_cached_value=True).get("jobName")

    @property
    def _logs_prefix(self) -> str:
        """str: the prefix for the job logs."""
        # jobs ARNs used to contain the job name and use a log prefix of `job-name`
        # now job ARNs use a UUID and a log prefix of `job-name/UUID`
        return (
            f"{self.name}"
            if self.arn.endswith(self.name)
            else f"{self.name}/{self.arn.split('/')[-1]}"
        )

    def state(self, use_cached_value: bool = False) -> str:
        """The state of the quantum hybrid job.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved
                value from the Amazon Braket `GetJob` operation. If `False`, calls the
                `GetJob` operation to retrieve metadata, which also updates the cached
                value. Default = `False`.

        Returns:
            str: The value of `status` in `metadata()`. This is the value of the `status` key
            in the Amazon Braket `GetJob` operation.

        See Also:
            `metadata()`
        """
        return self.metadata(use_cached_value).get("status")

    def queue_position(self) -> HybridJobQueueInfo:
        """The queue position details for the hybrid job.

        Returns:
            HybridJobQueueInfo: Instance of HybridJobQueueInfo class representing
            the queue position information for the hybrid job. The queue_position is
            only returned when the hybrid job is not in RUNNING/CANCELLING/TERMINAL states,
            else queue_position is returned as None. If the queue position of the hybrid
            job is greater than 15, we return '>15' as the queue_position return value.

        Examples:
            job status = QUEUED and position is 2 in the queue.
            >>> job.queue_position()
            HybridJobQueueInfo(queue_position='2', message=None)

            job status = QUEUED and position is 18 in the queue.
            >>> job.queue_position()
            HybridJobQueueInfo(queue_position='>15', message=None)

            job status = COMPLETED
            >>> job.queue_position()
            HybridJobQueueInfo(queue_position=None,
            message='Job is in COMPLETED status. AmazonBraket does
                        not show queue position for this status.')
        """
        response = self.metadata()["queueInfo"]
        queue_position = None if response.get("position") == "None" else response.get("position")
        message = response.get("message")

        return HybridJobQueueInfo(queue_position=queue_position, message=message)

    def logs(self, wait: bool = False, poll_interval_seconds: int = 5) -> None:
        """Display logs for a given hybrid job, optionally tailing them until hybrid job is
           complete.

        If the output is a tty or a Jupyter cell, it will be color-coded
        based on which instance the log entry is from.

        Args:
            wait (bool): `True` to keep looking for new log entries until the hybrid job completes;
                otherwise `False`. Default: `False`.

            poll_interval_seconds (int): The interval of time, in seconds, between polling for
                new log entries and hybrid job completion (default: 5).

        Raises:
            exceptions.UnexpectedStatusException: If waiting and the training hybrid job fails.
        """
        # The loop below implements a state machine that alternates between checking the hybrid job
        # status and reading whatever is available in the logs at this point. Note, that if we were
        # called with wait == False, we never check the hybrid job status.
        #
        # If wait == TRUE and hybrid job is not completed, the initial state is TAILING
        # If wait == FALSE, the initial state is COMPLETE (doesn't matter if the hybrid job really
        # is complete).
        #
        # The state table:
        #
        # STATE               ACTIONS                        CONDITION             NEW STATE
        # ----------------    ----------------               -----------------     ----------------
        # TAILING             Read logs, Pause, Get status   Job complete          JOB_COMPLETE
        #                                                    Else                  TAILING
        # JOB_COMPLETE        Read logs, Pause               Any                   COMPLETE
        # COMPLETE            Read logs, Exit                                      N/A
        #
        # Notes:
        # - The JOB_COMPLETE state forces us to do an extra pause and read any items that got to
        #   Cloudwatch after the job was marked complete.

        job_already_completed = self.state() in AwsQuantumJob.TERMINAL_STATES
        log_state = (
            AwsQuantumJob.LogState.TAILING
            if wait and not job_already_completed
            else AwsQuantumJob.LogState.COMPLETE
        )

        log_group = AwsQuantumJob.LOG_GROUP
        stream_names = []  # The list of log streams
        positions = {}  # The current position in each stream, map of stream name -> position
        instance_count = self.metadata(use_cached_value=True)["instanceConfig"]["instanceCount"]
        has_streams = False
        color_wrap = logs.ColorWrap()
        previous_state = self.state()

        while True:
            time.sleep(poll_interval_seconds)
            current_state = self.state()
            has_streams = logs.flush_log_streams(
                self._aws_session,
                log_group,
                self._logs_prefix,
                stream_names,
                positions,
                instance_count,
                has_streams,
                color_wrap,
                [previous_state, current_state],
                self.queue_position().queue_position if not self._quiet else None,
            )
            previous_state = current_state

            if log_state == AwsQuantumJob.LogState.COMPLETE:
                break

            if log_state == AwsQuantumJob.LogState.JOB_COMPLETE:
                log_state = AwsQuantumJob.LogState.COMPLETE
            elif current_state in AwsQuantumJob.TERMINAL_STATES:
                log_state = AwsQuantumJob.LogState.JOB_COMPLETE

    def metadata(self, use_cached_value: bool = False) -> dict[str, Any]:
        """Gets the hybrid job metadata defined in Amazon Braket.

        Args:
            use_cached_value (bool): If `True`, uses the value most recently retrieved
                from the Amazon Braket `GetJob` operation, if it exists; if does not exist,
                `GetJob` is called to retrieve the metadata. If `False`, always calls
                `GetJob`, which also updates the cached value. Default: `False`.

        Returns:
            dict[str, Any]: Dict that specifies the hybrid job metadata defined in Amazon Braket.
        """
        if not use_cached_value or not self._metadata:
            self._metadata = self._aws_session.get_job(self._arn)
        return self._metadata

    def metrics(
        self,
        metric_type: MetricType = MetricType.TIMESTAMP,
        statistic: MetricStatistic = MetricStatistic.MAX,
    ) -> dict[str, list[Any]]:
        """Gets all the metrics data, where the keys are the column names, and the values are a list
        containing the values in each row. For example, the table:
            timestamp energy
              0         0.1
              1         0.2
        would be represented as:
        { "timestamp" : [0, 1], "energy" : [0.1, 0.2] }
        values may be integers, floats, strings or None.

        Args:
            metric_type (MetricType): The type of metrics to get. Default: MetricType.TIMESTAMP.

            statistic (MetricStatistic): The statistic to determine which metric value to use
                when there is a conflict. Default: MetricStatistic.MAX.

        Returns:
            dict[str, list[Any]]: The metrics data.
        """
        fetcher = CwlInsightsMetricsFetcher(self._aws_session)
        metadata = self.metadata(True)
        job_start = None
        job_end = None
        if "startedAt" in metadata:
            job_start = int(metadata["startedAt"].timestamp())
        if self.state() in AwsQuantumJob.TERMINAL_STATES and "endedAt" in metadata:
            job_end = int(math.ceil(metadata["endedAt"].timestamp()))
        return fetcher.get_metrics_for_job(
            self.name, metric_type, statistic, job_start, job_end, self._logs_prefix
        )

    def cancel(self) -> str:
        """Cancels the job.

        Returns:
            str: Indicates the status of the job.

        Raises:
            ClientError: If there are errors invoking the CancelJob API.
        """
        cancellation_response = self._aws_session.cancel_job(self._arn)
        return cancellation_response["cancellationStatus"]

    def result(
        self,
        poll_timeout_seconds: float = QuantumJob.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = QuantumJob.DEFAULT_RESULTS_POLL_INTERVAL,
    ) -> dict[str, Any]:
        """Retrieves the hybrid job result persisted using the `save_job_result` function.

        Args:
            poll_timeout_seconds (float): The polling timeout, in seconds, for `result()`.
                Default: 10 days.
            poll_interval_seconds (float): The polling interval, in seconds, for `result()`.
                Default: 5 seconds.

        Returns:
            dict[str, Any]: Dict specifying the job results.

        Raises:
            RuntimeError: if hybrid job is in a FAILED or CANCELLED state.
            TimeoutError: if hybrid job execution exceeds the polling timeout period.
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            job_name = self.metadata(True)["jobName"]

            try:
                self.download_result(temp_dir, poll_timeout_seconds, poll_interval_seconds)
            except ClientError as e:
                if e.response["Error"]["Code"] == "404":
                    return {}
                else:
                    raise e
            return AwsQuantumJob._read_and_deserialize_results(temp_dir, job_name)

    @staticmethod
    def _read_and_deserialize_results(temp_dir: str, job_name: str) -> dict[str, Any]:
        return load_job_result(Path(temp_dir, job_name, AwsQuantumJob.RESULTS_FILENAME))

    def download_result(
        self,
        extract_to: str | None = None,
        poll_timeout_seconds: float = QuantumJob.DEFAULT_RESULTS_POLL_TIMEOUT,
        poll_interval_seconds: float = QuantumJob.DEFAULT_RESULTS_POLL_INTERVAL,
    ) -> None:
        """Downloads the results from the hybrid job output S3 bucket and extracts the tar.gz
        bundle to the location specified by `extract_to`. If no location is specified,
        the results are extracted to the current directory.

        Args:
            extract_to (str | None): The directory to which the results are extracted. The results
                are extracted to a folder titled with the hybrid job name within this directory.
                Default= `Current working directory`.
            poll_timeout_seconds (float): The polling timeout, in seconds, for `download_result()`.
                Default: 10 days.
            poll_interval_seconds (float): The polling interval, in seconds, for
                `download_result()`.Default: 5 seconds.

        Raises:
            RuntimeError: if hybrid job is in a FAILED or CANCELLED state.
            TimeoutError: if hybrid job execution exceeds the polling timeout period.
        """
        extract_to = extract_to or Path.cwd()

        timeout_time = time.time() + poll_timeout_seconds
        job_response = self.metadata(True)

        while time.time() < timeout_time:
            job_response = self.metadata(True)
            job_state = self.state()

            if job_state in AwsQuantumJob.TERMINAL_STATES:
                output_s3_path = job_response["outputDataConfig"]["s3Path"]
                output_s3_uri = f"{output_s3_path}/output/model.tar.gz"
                AwsQuantumJob._attempt_results_download(self, output_s3_uri, output_s3_path)
                AwsQuantumJob._extract_tar_file(f"{extract_to}/{self.name}")
                return
            else:
                time.sleep(poll_interval_seconds)

        raise TimeoutError(
            f"{job_response['jobName']}: Polling for job completion "
            f"timed out after {poll_timeout_seconds} seconds."
        )

    def _attempt_results_download(self, output_bucket_uri: str, output_s3_path: str) -> None:
        try:
            self._aws_session.download_from_s3(
                s3_uri=output_bucket_uri, filename=AwsQuantumJob.RESULTS_TAR_FILENAME
            )
        except ClientError as e:
            if e.response["Error"]["Code"] == "404":
                exception_response = {
                    "Error": {
                        "Code": "404",
                        "Message": f"Error retrieving results, "
                        f"could not find results at '{output_s3_path}'",
                    }
                }
                raise ClientError(exception_response, "HeadObject") from e
            else:
                raise e

    @staticmethod
    def _extract_tar_file(extract_path: str) -> None:
        with tarfile.open(AwsQuantumJob.RESULTS_TAR_FILENAME, "r:gz") as tar:
            tar.extractall(extract_path)

    def __repr__(self) -> str:
        return f"AwsQuantumJob('arn':'{self.arn}')"

    def __eq__(self, other: AwsQuantumJob) -> bool:
        if isinstance(other, AwsQuantumJob):
            return self.arn == other.arn
        return False

    def __hash__(self) -> int:
        return hash(self.arn)

    @staticmethod
    def _initialize_session(session_value: AwsSession, device: str, logger: Logger) -> AwsSession:
        aws_session = session_value or AwsSession()
        if device.startswith("local:"):
            return aws_session
        device_region = AwsDevice.get_device_region(device)
        return (
            AwsQuantumJob._initialize_regional_device_session(aws_session, device, logger)
            if device_region
            else AwsQuantumJob._initialize_non_regional_device_session(aws_session, device, logger)
        )

    @staticmethod
    def _initialize_regional_device_session(
        aws_session: AwsSession, device: AwsDevice, logger: Logger
    ) -> AwsSession:
        device_region = AwsDevice.get_device_region(device)
        current_region = aws_session.region
        if current_region != device_region:
            aws_session = aws_session.copy_session(region=device_region)
            logger.info(f"Changed session region from '{current_region}' to '{device_region}'")
        try:
            aws_session.get_device(device)
            return aws_session
        except ClientError as e:
            raise (
                ValueError(f"'{device}' not found.")
                if e.response["Error"]["Code"] == "ResourceNotFoundException"
                else e
            )

    @staticmethod
    def _initialize_non_regional_device_session(
        aws_session: AwsSession, device: AwsDevice, logger: Logger
    ) -> AwsSession:
        original_region = aws_session.region
        try:
            aws_session.get_device(device)
            return aws_session
        except ClientError as e:
            if e.response["Error"]["Code"] == "ResourceNotFoundException":
                if "qpu" not in device:
                    raise ValueError(f"Simulator '{device}' not found in '{original_region}'")
            else:
                raise e

        for region in frozenset(AwsDevice.REGIONS) - {original_region}:
            device_session = aws_session.copy_session(region=region)
            try:
                device_session.get_device(device)
                logger.info(
                    f"Changed session region from '{original_region}' to '{device_session.region}'"
                )
                return device_session
            except ClientError as e:
                if e.response["Error"]["Code"] != "ResourceNotFoundException":
                    raise e
        raise ValueError(f"QPU '{device}' not found.")



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import uuid

import numpy as np

from braket.task_result import TaskMetadata
from braket.tasks import GateModelQuantumTaskResult
from braket.tasks.local_quantum_task_batch import LocalQuantumTaskBatch

RESULTS = [
    GateModelQuantumTaskResult(
        task_metadata=TaskMetadata(
            **{"id": str(uuid.uuid4()), "deviceId": "default", "shots": 100}
        ),
        additional_metadata=None,
        measurements=np.array([[0, 1], [1, 0]]),
        measured_qubits=[0, 1],
        result_types=None,
        values=None,
    )
]

TASK_BATCH = LocalQuantumTaskBatch(RESULTS)


def test_results():
    assert TASK_BATCH.results() == RESULTS
    assert len(TASK_BATCH.results()) == 1



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterable
from enum import Enum
from typing import Any, Union


class CriteriaKey(str, Enum):
    """Specifies the types of keys that a criteria may use to match an instruction,
    observable, etc.
    """

    QUBIT = "QUBIT"
    GATE = "GATE"
    UNITARY_GATE = "UNITARY_GATE"
    OBSERVABLE = "OBSERVABLE"


class CriteriaKeyResult(str, Enum):
    """The get_keys() method may return this enum instead of actual keys for
    a given criteria key type.
    """

    ALL = "ALL"


class Criteria(ABC):
    """Represents conditions that need to be met for a noise to apply to a circuit."""

    @abstractmethod
    def applicable_key_types(self) -> Iterable[CriteriaKey]:
        """Returns the relevant set of keys for the Criteria

        This informs what the Criteria operates on and can be used to optimize
        which Criteria is relevant.

        Returns:
            Iterable[CriteriaKey]: The relevant set of keys for the Criteria.
        """
        raise NotImplementedError

    @abstractmethod
    def get_keys(self, key_type: CriteriaKey) -> Union[CriteriaKeyResult, set[Any]]:
        """Returns a set of key for a given key type.

        Args:
            key_type (CriteriaKey): The criteria key type.

        Returns:
            Union[CriteriaKeyResult, set[Any]]: Returns a set of keys for a key type. The
            actual returned keys will depend on the CriteriaKey. If the provided key type
            is not relevant the returned list will be empty. If the provided key type is
            relevant for all possible inputs, the string CriteriaKeyResult.ALL will be returned.
        """
        raise NotImplementedError

    def __eq__(self, other: Criteria):
        if not isinstance(other, Criteria):
            return NotImplemented
        if self.applicable_key_types() != other.applicable_key_types():
            return False
        for key_type in self.applicable_key_types():
            if self.get_keys(key_type) != other.get_keys(key_type):
                return False
        return True

    @abstractmethod
    def to_dict(self) -> dict:
        """Converts this Criteria object into a dict representation

        Returns:
            dict: A dictionary object representing the Criteria.
        """
        raise NotImplementedError

    @classmethod
    def from_dict(cls, criteria: dict) -> Criteria:
        """Converts a dictionary representing an object of this class into an instance of this
        class.

        Args:
            criteria (dict): A dictionary representation of an object of this class.

        Returns:
            Criteria: An object of this class that corresponds to the passed in dictionary.
        """
        if "__class__" in criteria:
            criteria_name = criteria["__class__"]
            criteria_cls = getattr(Criteria, criteria_name)
            return criteria_cls.from_dict(criteria)
        raise NotImplementedError

    @classmethod
    def register_criteria(cls, criteria: type[Criteria]) -> None:
        """Register a criteria implementation by adding it into the Criteria class.

        Args:
            criteria (type[Criteria]): Criteria class to register.
        """
        setattr(cls, criteria.__name__, criteria)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

from dataclasses import dataclass

import numpy as np

from braket.task_result import AdditionalMetadata, PhotonicModelTaskResult, TaskMetadata


@dataclass
class PhotonicModelQuantumTaskResult:
    task_metadata: TaskMetadata
    additional_metadata: AdditionalMetadata
    measurements: np.ndarray = None

    def __eq__(self, other: PhotonicModelQuantumTaskResult) -> bool:
        if isinstance(other, PhotonicModelQuantumTaskResult):
            return self.task_metadata.id == other.task_metadata.id
        return NotImplemented

    @staticmethod
    def from_object(result: PhotonicModelTaskResult) -> PhotonicModelQuantumTaskResult:
        """Create PhotonicModelQuantumTaskResult from PhotonicModelTaskResult object.

        Args:
            result (PhotonicModelTaskResult): PhotonicModelTaskResult object

        Returns:
            PhotonicModelQuantumTaskResult: A PhotonicModelQuantumTaskResult based on the given dict

        Raises:
            ValueError: If "measurements" is not a key in the result dict
        """
        return PhotonicModelQuantumTaskResult._from_object_internal(result)

    @staticmethod
    def from_string(result: str) -> PhotonicModelQuantumTaskResult:
        return PhotonicModelQuantumTaskResult._from_object_internal(
            PhotonicModelTaskResult.parse_raw(result)
        )

    @classmethod
    def _from_object_internal(
        cls, result: PhotonicModelTaskResult
    ) -> PhotonicModelQuantumTaskResult:
        task_metadata = result.taskMetadata
        additional_metadata = result.additionalMetadata
        if result.measurements is not None:
            measurements = np.asarray(result.measurements, dtype=int)
        else:
            measurements = None
        return cls(
            task_metadata=task_metadata,
            additional_metadata=additional_metadata,
            measurements=measurements,
        )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import unittest.mock as mock
from datetime import timedelta

import boto3
import pytest
from botocore.exceptions import ClientError

from braket.aws import AwsDevice, AwsSession
from braket.circuits import Circuit
from braket.tracking import Tracker
from braket.tracking.tracker import MIN_SIMULATOR_DURATION


@pytest.mark.parametrize(
    "qpu",
    [
        "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony",
        "arn:aws:braket:eu-west-2::device/qpu/oqc/Lucy",
        "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3",
    ],
)
def test_qpu_tracking(qpu):
    circuit = Circuit().h(0)
    with Tracker() as t:
        device = AwsDevice(qpu)
        # Mock out task creation against the service
        device._aws_session.braket_client.create_quantum_task = mock.Mock(
            return_value={
                "quantumTaskArn": (
                    f"arn:aws:braket:{device._aws_session.region}"
                    ":1234567890:quantum-task/e9e6bd31-5ba3-4027-948d-93c5f12e2942"
                )
            }
        )
        device.run(circuit, shots=10)

    assert t.qpu_tasks_cost() > 0


def test_simulator_tracking():
    circuit = Circuit().h(0).cnot(0, 1)
    device = AwsDevice("arn:aws:braket:::device/quantum-simulator/amazon/sv1")

    with Tracker() as t:
        task0 = device.run(circuit, shots=100)
        task1 = device.run(circuit, shots=100)
        assert t.quantum_tasks_statistics() == {
            "arn:aws:braket:::device/quantum-simulator/amazon/sv1": {
                "shots": 200,
                "tasks": {"CREATED": 2},
            }
        }
        task0.result()
        task1.result()

        try:
            device.run(circuit, shots=100).cancel()
        except ClientError as e:
            if not e.response["Error"]["Message"].startswith(
                "Amazon Braket cannot cancel a quantum task in the COMPLETED status"
            ):
                raise e

    quantum_stats = t.quantum_tasks_statistics()[device.arn]
    assert quantum_stats["shots"] == 300
    assert quantum_stats["tasks"] == {"COMPLETED": 2, "CANCELLING": 1}
    assert quantum_stats["execution_duration"] > timedelta(0)
    assert quantum_stats["billed_execution_duration"] >= quantum_stats["execution_duration"]
    assert quantum_stats["billed_execution_duration"] >= 2 * MIN_SIMULATOR_DURATION

    assert t.qpu_tasks_cost() == 0
    assert t.simulator_tasks_cost() > 0


def test_all_devices_price_search():
    devices = AwsDevice.get_devices(statuses=["ONLINE", "OFFLINE"])

    tasks = {}
    for region in AwsDevice.REGIONS:
        s = AwsSession(boto3.Session(region_name=region))
        # Skip devices with empty execution windows
        for device in [device for device in devices if device.properties.service.executionWindows]:
            try:
                s.get_device(device.arn)

                # If we are here, device can create tasks in region
                details = {
                    "shots": 100,
                    "device": device.arn,
                    "billed_duration": MIN_SIMULATOR_DURATION,
                    "job_task": False,
                    "status": "COMPLETED",
                }
                tasks[f"task:for:{device.name}:{region}"] = details.copy()
                details["job_task"] = True
                tasks[f"jobtask:for:{device.name}:{region}"] = details
            except s.braket_client.exceptions.ResourceNotFoundException:
                # device does not exist in region, so nothing to test
                pass

    t = Tracker()
    t._resources = tasks

    assert t.qpu_tasks_cost() + t.simulator_tasks_cost() > 0



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from __future__ import annotations

import csv
import io
import os
from functools import lru_cache

import urllib3


class Pricing:
    def __init__(self):
        self._price_list = []

    def get_prices(self) -> None:
        """Retrieves the price list."""
        # Using AWS Pricing Bulk API. Format for the response is described at
        # https://docs.aws.amazon.com/awsaccountbilling/latest/aboutv2/reading-an-offer.html

        http = urllib3.PoolManager()
        price_url = os.environ.get(
            "BRAKET_PRICE_OFFERS_URL",
            "https://pricing.us-east-1.amazonaws.com/offers/v1.0/aws/AmazonBraket/current/index.csv",  # noqa: E501
        )
        response = http.request(
            "GET",
            price_url,
            preload_content=False,
        )
        response.auto_close = False

        text_response = io.TextIOWrapper(response)

        # Data starts on line 6
        #
        # > The first five rows of the CSV are the metadata for the offer file. The sixth row has
        # > all the column names for the products and their attributes...
        # https://docs.aws.amazon.com/awsaccountbilling/latest/aboutv2/reading-an-offer.html#csv
        for _ in range(5):
            text_response.readline()
        self._price_list = list(csv.DictReader(text_response))

    @lru_cache
    def price_search(self, **kwargs: str) -> list[dict[str, str]]:
        """Searches the price list for a given set of parameters.

        Args:
            **kwargs (str): Arbitrary keyword arguments.

        Returns:
            list[dict[str, str]]: The price list.
        """
        if not self._price_list:
            self.get_prices()
        return [
            entry for entry in self._price_list if all(entry[k] == v for k, v in kwargs.items())
        ]


_pricing = Pricing()
price_search = _pricing.price_search



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from collections.abc import Iterable
from typing import Any, Optional, Union

from braket.circuits.noise_model.criteria import Criteria, CriteriaKey, CriteriaKeyResult
from braket.circuits.noise_model.criteria_input_parsing import parse_qubit_input
from braket.circuits.noise_model.initialization_criteria import InitializationCriteria
from braket.registers.qubit_set import QubitSet, QubitSetInput


class QubitInitializationCriteria(InitializationCriteria):
    """This class models initialization noise Criteria based on qubits."""

    def __init__(self, qubits: Optional[QubitSetInput] = None):
        """Creates initialization noise Qubit-based Criteria.

        Args:
            qubits (Optional[QubitSetInput]): A set of relevant qubits. If no qubits
                are provided, all (possible) qubits are considered to be relevant.
        """
        self._qubits = parse_qubit_input(qubits)

    def __str__(self):
        return f"{self.__class__.__name__}({self._qubits})"

    def __repr__(self):
        return f"{self.__class__.__name__}(qubits={self._qubits})"

    def applicable_key_types(self) -> Iterable[CriteriaKey]:
        """Gets the QUBIT criteria key.

        Returns:
            Iterable[CriteriaKey]: This Criteria operates on Qubits, but is valid for all Gates.
        """
        return [CriteriaKey.QUBIT]

    def get_keys(self, key_type: CriteriaKey) -> Union[CriteriaKeyResult, set[Any]]:
        """Gets the keys for a given CriteriaKey.

        Args:
            key_type (CriteriaKey): The relevant Criteria Key.

        Returns:
            Union[CriteriaKeyResult, set[Any]]: The return value is based on the key type:
            QUBIT will return a set of qubit targets that are relevant to this Critera, or
            CriteriaKeyResult.ALL if the Criteria is relevant for all (possible) qubits.
            All other keys will return an empty set.
        """
        if key_type == CriteriaKey.QUBIT:
            if self._qubits is None:
                return CriteriaKeyResult.ALL
            return set(self._qubits)
        return set()

    def to_dict(self) -> dict:
        """Converts a dictionary representing an object of this class into an instance of
        this class.

        Returns:
            dict: A dictionary representing the serialized version of this Criteria.
        """
        qubits = list(self._qubits) if self._qubits is not None else None
        return {
            "__class__": self.__class__.__name__,
            "qubits": qubits,
        }

    def qubit_intersection(self, qubits: QubitSetInput) -> QubitSetInput:
        """Returns subset of passed qubits that match the criteria.

        Args:
            qubits (QubitSetInput): A qubit or set of qubits that may match the criteria.

        Returns:
            QubitSetInput: The subset of passed qubits that match the criteria.
        """
        target_qubit = QubitSet(qubits)
        if self._qubits is None:
            return target_qubit
        return self._qubits.intersection(target_qubit)

    @classmethod
    def from_dict(cls, criteria: dict) -> Criteria:
        """Deserializes a dictionary into a Criteria object.

        Args:
            criteria (dict): A dictionary representation of a QubitCriteria.

        Returns:
            Criteria: A deserialized QubitCriteria represented by the passed in
            serialized data.
        """
        return QubitInitializationCriteria(criteria["qubits"])


Criteria.register_criteria(QubitInitializationCriteria)



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from abc import abstractmethod

from braket.circuits.noise_model.criteria import Criteria
from braket.circuits.result_type import ResultType


class ResultTypeCriteria(Criteria):
    """Criteria that implement these methods may be used to determine readout noise."""

    @abstractmethod
    def result_type_matches(self, result_type: ResultType) -> bool:
        """Returns true if a result type matches the criteria.

        Args:
            result_type (ResultType): A result type or list of result types to match.

        Returns:
            bool: True if the result type matches the criteria.
        """
        raise NotImplementedError



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Union

from braket.tasks import (
    AnnealingQuantumTaskResult,
    GateModelQuantumTaskResult,
    PhotonicModelQuantumTaskResult,
    QuantumTaskBatch,
)


class LocalQuantumTaskBatch(QuantumTaskBatch):
    """Executes a batch of quantum tasks in parallel.

    Since this class is instantiated with the results, cancel() and run_async() are unsupported.
    """

    def __init__(
        self,
        results: list[
            Union[
                GateModelQuantumTaskResult,
                AnnealingQuantumTaskResult,
                PhotonicModelQuantumTaskResult,
            ]
        ],
    ):
        self._results = results

    def results(
        self,
    ) -> list[
        Union[
            GateModelQuantumTaskResult, AnnealingQuantumTaskResult, PhotonicModelQuantumTaskResult
        ]
    ]:
        return self._results



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import re

import pytest

from braket.circuits import Observable, ObservableResultType, ResultType
from braket.circuits.free_parameter import FreeParameter
from braket.circuits.result_type import ObservableParameterResultType
from braket.circuits.serialization import IRType


@pytest.fixture
def result_type():
    return ResultType(ascii_symbols=["foo"])


@pytest.fixture
def prob():
    return ResultType.Probability([0, 1])


@pytest.fixture
def sv():
    return ResultType.StateVector()


def test_none_ascii():
    with pytest.raises(ValueError):
        ResultType(ascii_symbols=None)


def test_name(result_type):
    expected = result_type.__class__.__name__
    assert result_type.name == expected


def test_ascii_symbol():
    ascii_symbols = ["foo"]
    result_type = ResultType(ascii_symbols=ascii_symbols)
    assert result_type.ascii_symbols == ascii_symbols


def test_equality_statevector():
    result1 = ResultType.StateVector()
    result2 = ResultType.StateVector()
    result3 = ResultType.Probability([1])
    result4 = "hi"
    assert result1 == result2
    assert result1 != result3
    assert result1 != result4


def test_equality_densitymatrix():
    result1 = ResultType.DensityMatrix()
    result2 = ResultType.DensityMatrix()
    result3 = ResultType.StateVector()
    result4 = "foo"
    assert result1 == result2
    assert result1 != result3
    assert result1 != result4


def test_ascii_symbol_setter(result_type):
    with pytest.raises(AttributeError):
        result_type.ascii_symbols = ["bar"]


def test_name_setter(result_type):
    with pytest.raises(AttributeError):
        result_type.name = "hi"


def test_register_result():
    class _FooResultType(ResultType):
        def __init__(self):
            super().__init__(ascii_symbols=["foo"])

    ResultType.register_result_type(_FooResultType)
    assert ResultType._FooResultType().name == _FooResultType().name


def test_copy_creates_new_object(prob):
    copy = prob.copy()
    assert copy == prob
    assert copy is not prob


def test_copy_with_mapping_target(sv):
    target_mapping = {0: 10, 1: 11}
    expected = ResultType.StateVector()
    assert sv.copy(target_mapping=target_mapping) == expected


def test_copy_with_mapping_target_hasattr(prob):
    target_mapping = {0: 10, 1: 11}
    expected = ResultType.Probability([10, 11])
    assert prob.copy(target_mapping=target_mapping) == expected


def test_copy_with_target_hasattr(prob):
    target = [10, 11]
    expected = ResultType.Probability(target)
    assert prob.copy(target=target) == expected


def test_copy_with_target(sv):
    target = [10, 11]
    expected = ResultType.StateVector()
    assert sv.copy(target=target) == expected


def test_copy_with_target_and_mapping(prob):
    with pytest.raises(TypeError):
        prob.copy(target=[10], target_mapping={0: 10})


# ObservableResultType


def test_expectation_init_value_error_target():
    with pytest.raises(ValueError):
        ObservableResultType(
            ascii_symbols=["Obs", "Obs"], observable=Observable.X() @ Observable.Y(), target=[]
        )


def test_expectation_init_value_error_ascii_symbols():
    with pytest.raises(ValueError):
        ObservableResultType(
            ascii_symbols=["Obs"], observable=Observable.X() @ Observable.Y(), target=[1, 2]
        )


def test_obs_rt_init_value_error_qubit_count():
    with pytest.raises(ValueError):
        ObservableResultType(ascii_symbols=["Obs"], observable=Observable.X(), target=[0, 1])


def test_obs_rt_equality():
    a1 = ObservableResultType(ascii_symbols=["Obs"], observable=Observable.X(), target=0)
    a2 = ObservableResultType(ascii_symbols=["Obs"], observable=Observable.X(), target=0)
    a3 = ObservableResultType(ascii_symbols=["Obs"], observable=Observable.X(), target=1)
    a4 = "hi"
    assert a1 == a2
    assert a1 != a3
    assert a1 != a4
    assert ResultType.Variance(observable=Observable.Y(), target=0) != ResultType.Expectation(
        observable=Observable.Y(), target=0
    )


def test_obs_rt_repr():
    a1 = ObservableResultType(ascii_symbols=["Obs"], observable=Observable.X(), target=0)
    assert (
        str(a1)
        == "ObservableResultType(observable=X('qubit_count': 1), target=QubitSet([Qubit(0)]))"
    )


@pytest.mark.parametrize(
    "ir_type, serialization_properties, expected_exception, expected_message",
    [
        (IRType.JAQCD, None, NotImplementedError, "to_jaqcd has not been implemented yet."),
        (IRType.OPENQASM, None, NotImplementedError, "to_openqasm has not been implemented yet."),
        ("invalid-ir-type", None, ValueError, "Supplied ir_type invalid-ir-type is not supported."),
        (
            IRType.OPENQASM,
            "invalid-serialization-properties",
            ValueError,
            "serialization_properties must be of type OpenQASMSerializationProperties for "
            "IRType.OPENQASM.",
        ),
    ],
)
def test_result_type_to_ir(
    ir_type, serialization_properties, expected_exception, expected_message, result_type
):
    with pytest.raises(expected_exception) as exc:
        result_type.to_ir(ir_type, serialization_properties=serialization_properties)
    assert exc.value.args[0] == expected_message


# Observable Result Type with Params


def test_expectation_init_value_error_target_adjoint_gradient():
    tensor_operation_error = re.escape(
        "Observable TensorProduct(X('qubit_count': 1), "
        "Y('qubit_count': 1)) must only operate on 1 qubit for target=None"
    )
    with pytest.raises(ValueError, match=tensor_operation_error):
        ObservableParameterResultType(
            ascii_symbols=["Obs", "Obs"],
            observable=Observable.X() @ Observable.Y(),
            target=[],
            parameters=["alpha"],
        )


def test_expectation_init_value_error_ascii_symbols_adjoint_gradient():
    ascii_and_obs_qubit_count_mismatch = (
        "Observable's qubit count and the number of ASCII symbols must be equal"
    )
    with pytest.raises(ValueError, match=ascii_and_obs_qubit_count_mismatch):
        ObservableParameterResultType(
            ascii_symbols=["Obs"],
            observable=Observable.X() @ Observable.Y(),
            target=[1, 2],
            parameters=[],
        )


def test_obs_rt_init_value_error_qubit_count_adjoint_gradient():
    obs_and_target_count_mismatch = re.escape(
        "Observable's qubit count 1 and the size of the target "
        "qubit set QubitSet([Qubit(0), Qubit(1)]) must be equal"
    )
    with pytest.raises(ValueError, match=obs_and_target_count_mismatch):
        ObservableParameterResultType(
            ascii_symbols=["Obs"], observable=Observable.X(), target=[0, 1]
        )


def test_valid_result_type_for__adjoint_gradient():
    ObservableParameterResultType(
        ascii_symbols=["Obs", "Obs"],
        observable=Observable.X() @ Observable.Y(),
        target=[0, 1],
        parameters=["alpha", FreeParameter("beta")],
    )


def test_obs_rt_repr_adjoint_gradient():
    a1 = ObservableParameterResultType(
        ascii_symbols=["Obs"], observable=Observable.X(), target=0, parameters=["alpha"]
    )
    assert (
        str(a1) == "ObservableParameterResultType(observable=X('qubit_count': 1), "
        "target=QubitSet([Qubit(0)]), parameters=['alpha'])"
    )



# Copyright Amazon.com Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import os
import random
import string

import boto3
import pytest
from botocore.exceptions import ClientError

from braket.aws.aws_device import AwsDevice
from braket.aws.aws_quantum_job import AwsQuantumJob
from braket.aws.aws_session import AwsSession

SV1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/sv1"
DM1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/dm1"
TN1_ARN = "arn:aws:braket:::device/quantum-simulator/amazon/tn1"
SIMULATOR_ARNS = [SV1_ARN, DM1_ARN, TN1_ARN]

job_complete_name = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))
job_fail_name = "".join(random.choices(string.ascii_lowercase + string.digits, k=12))


def pytest_configure_node(node):
    """xdist hook"""
    node.workerinput["JOB_COMPLETED_NAME"] = job_complete_name
    node.workerinput["JOB_FAILED_NAME"] = job_fail_name
    if endpoint := os.getenv("BRAKET_ENDPOINT"):
        node.workerinput["BRAKET_ENDPOINT"] = endpoint


def pytest_xdist_node_collection_finished(ids):
    """Uses the pytest xdist hook to check whether tests with jobs are to be ran.
    If they are, the first reporting worker sets a flag that it created the tests
    to avoid concurrency limits. This is the first time in the pytest setup the
    controller has all the tests to be ran from the worker nodes.
    """
    run_jobs = any("job" in test for test in ids)
    profile_name = os.environ["AWS_PROFILE"]
    aws_session = AwsSession(boto3.session.Session(profile_name=profile_name))
    if run_jobs and os.getenv("JOBS_STARTED") is None:
        AwsQuantumJob.create(
            "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            job_name=job_fail_name,
            source_module="test/integ_tests/job_test_script.py",
            entry_point="job_test_script:start_here",
            aws_session=aws_session,
            wait_until_complete=False,
            hyperparameters={"test_case": "failed"},
        )
        AwsQuantumJob.create(
            "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            job_name=job_complete_name,
            source_module="test/integ_tests/job_test_script.py",
            entry_point="job_test_script:start_here",
            aws_session=aws_session,
            wait_until_complete=False,
            hyperparameters={"test_case": "completed"},
        )
        os.environ["JOBS_STARTED"] = "True"


@pytest.fixture(scope="session")
def boto_session():
    profile_name = os.environ["AWS_PROFILE"]
    return boto3.session.Session(profile_name=profile_name)


@pytest.fixture(scope="session")
def aws_session(boto_session):
    return AwsSession(boto_session)


@pytest.fixture(scope="session")
def s3_resource(boto_session):
    return boto_session.resource("s3")


@pytest.fixture(scope="session")
def s3_client(boto_session):
    return boto_session.client("s3")


@pytest.fixture(scope="session")
def account_id(boto_session):
    return boto_session.client("sts").get_caller_identity()["Account"]


@pytest.fixture(scope="session")
def s3_bucket(s3_resource, s3_client, account_id, boto_session):
    """Create / get S3 bucket for tests"""

    region_name = boto_session.region_name
    bucket_name = f"amazon-braket-sdk-integ-tests-{account_id}"
    bucket = s3_resource.Bucket(bucket_name)

    try:
        # Determine if bucket exists
        s3_client.head_bucket(Bucket=bucket_name)
    except ClientError as e:
        error_code = e.response["Error"]["Code"]
        if error_code == "404":
            bucket.create(
                ACL="private", CreateBucketConfiguration={"LocationConstraint": region_name}
            )

    return bucket_name


@pytest.fixture(scope="module")
def s3_prefix():
    """Returns the module path of the test, e.g. integ_tests/test_simulator_quantum_task"""

    # current test path, e.g. ...
    # test/integ_tests/test_simulator_quantum_task.py::test_simulator_quantum_task (setup)
    current_test_path = os.environ.get("PYTEST_CURRENT_TEST")

    # strip off the filename extension and test/
    return current_test_path.rsplit(".py")[0].replace("test/", "")


@pytest.fixture(scope="module")
def s3_destination_folder(s3_bucket, s3_prefix):
    return AwsSession.S3DestinationFolder(s3_bucket, s3_prefix)


@pytest.fixture(scope="session")
def braket_simulators(aws_session):
    return {
        simulator_arn: AwsDevice(simulator_arn, aws_session) for simulator_arn in SIMULATOR_ARNS
    }


@pytest.fixture(scope="session")
def braket_devices():
    return AwsDevice.get_devices(statuses=["RETIRED", "ONLINE", "OFFLINE"])


@pytest.fixture(scope="session", autouse=True)
def created_braket_devices(aws_session, braket_devices):
    return {device.arn: device for device in braket_devices}


@pytest.fixture(scope="session")
def job_completed_name(request):
    return request.config.workerinput["JOB_COMPLETED_NAME"]


@pytest.fixture(scope="session")
def job_failed_name(request):
    return request.config.workerinput["JOB_FAILED_NAME"]


@pytest.fixture(scope="session", autouse=True)
def completed_quantum_job(job_completed_name):
    job_arn = [
        job["jobArn"]
        for job in boto3.client("braket").search_jobs(filters=[])["jobs"]
        if job["jobName"] == job_completed_name
    ][0]

    return AwsQuantumJob(arn=job_arn)


@pytest.fixture(scope="session", autouse=True)
def failed_quantum_job(job_failed_name):
    job_arn = [
        job["jobArn"]
        for job in boto3.client("braket").search_jobs(filters=[])["jobs"]
        if job["jobName"] == job_failed_name
    ][0]

    return AwsQuantumJob(arn=job_arn)
